/*ConvPresTempEdenIoniz solve for current pressure, calls PressureChange, ConvTempEdenIonize */
/*ConvFail handle conergece failure */
#include "cddefines.h"
#include "phycon.h"
#include "prtzone.h"
#include "mappar.h"
#include "called.h"
#include "thermal.h"
#include "punt.h"
#include "cooling.h"
#include "heat.h"
#include "pressure.h"
#include "trace.h"
#include "converge.h"

/* ----------------------------------------------------------------------------- */

/*ConvFail handle conergece failure */
static void ConvFail(char chMode[]); /* chMode[5] */

void ConvPresTempEdenIoniz(void)
{
	long int loop,
		LoopMax=20;
	double Error;
	float OldChange;
	int lgPresOscil;

#	ifdef DEBUG_FUN
	fputs( "<+>ConvPresTempEdenIoniz()\n", debug_fp );
#	endif

	/* this will count number of times we call ionize in this zone */
	conv.nPres2Ioniz = 0;
	loop = 0;
	conv.lgConvPres = FALSE;

	/* this will be the limit, which we will increase if no oscillations occur */
	LoopMax = 20;
	OldChange = 0.;

	/* this will be flag to check for pressure oscillations */
	lgPresOscil = FALSE;

	/* >>chng 01 apr 01, add test for at least 2 loops to get better pressure convergence */
	while( (loop < LoopMax) && !conv.lgConvPres  )
	{
		/* change current densities of all constituents if necessary, 
		 * PressureChange evaluates lgPresOK, true if pressure is now ok */
		PressureChange();

		/* heating cooling balance while doing ionization,
		 * this is where the heavy lifting is done*/
		ConvTempEdenIoniz();

		/* check whether pressure is oscillating */
		if( (pressure.pnow - pressure.pcorec ) * OldChange < 0. )
		{
			/* the sign of the change in pressure has changed, so things
			 * are oscillating.  This would be a problem */
			lgPresOscil = TRUE;
		}
		OldChange = pressure.pnow - pressure.pcorec ;

		/* convergence trace at this level */
		if( trace.lgTrConvg>=1  )
		{
			fprintf( ioQQQ, " ConvPresTempEdenIoniz%4ld Curr Pres:%10.2e Corr Pres:%10.2e Hden:%10.2e\n", 
			  loop, pressure.pnow, pressure.pcorec, phycon.hden );
		}

		/* >>chng 01 apr 02, if not constant density, then make sure at least two loops are done */
		if( (strcmp( pressure.chCPres,"CDEN") != 0 ) && (loop  == 1 ) )
			conv.lgConvPres = FALSE;

		/* increment loop counter */
		++loop;

		/* if we hit limit of loop, but no oscillations have happened, then we are
		 * making progress, and can keep going */
		if( loop == LoopMax && !lgPresOscil )
		{
			LoopMax = MIN2( 100 , LoopMax*2 );
		}
	}

	/* >>chng 01 mar 14, only call ConvFail one time, no matter how
	 * many failures occured below.  Had been series of if, so multiple
	 * calls per failure possible.  This change means that logic in ConvFail
	 * for abort is simpler. */
	if( !conv.lgConvPres )
	{
		/* announce the pressure failure */
		ConvFail("pres");
	}

	else if( !conv.lgConvTemp )
	{
		/* announce the temperature failure */
		ConvFail("temp");
	}

	/* say something if we did not converge */
	else if( !conv.lgConvEden )
	{
		ConvFail("eden") ;
	}

	/* say something if we did not converge */
	else if( !conv.lgConvIoniz )
	{
		ConvFail("ioni") ;
	}

	/* remember mean and largest errors on electron density */
	Error = fabs(phycon.eden - phycon.EdenTrue)/phycon.EdenTrue;
	if( Error > conv.BigEdenError )
	{
		conv.BigEdenError = (float)Error;
		phycon.nzEdenBad = nzone;
	}
	conv.AverEdenError += (float)Error;

	/* remember mean and largest errors between heating and cooling */
	Error = fabs(cooling.ctot - heat.htot) / cooling.ctot;
	conv.BigHeatCoolError = MAX2((float)Error , conv.BigHeatCoolError );
	conv.AverHeatCoolError += (float)Error;

	/* remember mean and largest pressure errors */
	Error = fabs(pressure.pnow - pressure.pcorec) / pressure.pcorec;
	conv.BigPressError = MAX2((float)Error , conv.BigPressError );
	conv.AverPressError += (float)Error;

#	ifdef DEBUG_FUN
	fputs( " <->ConvPresTempEdenIoniz()\n", debug_fp );
#	endif
	return;
}

/* ----------------------------------------------------------------------------- */

/*ConvFail handle conergece failure */
static void ConvFail(char chMode[]) /* chMode[5] */
{
	double relerror ;

#	ifdef DEBUG_FUN
	fputs( "<+>ConvFail()\n", debug_fp );
#	endif

	/* pressure failure */
	if( strcmp( chMode , "pres" )==0 )
	{
		/* record number of pressure failures */
		++conv.nPreFail;
		if( called.lgTalk )
		{
			fprintf( ioQQQ, 
				" convfail: pressure not converged; zone%4ld Te:%11.2e Hden:%10.2e curr Pres:%10.2e Corr Pres%10.2e\n", 
			  nzone, 
			  phycon.te, 
			  phycon.hden, 
			  pressure.pnow, 
			  pressure.pcorec );
		}
	}

	/* electron ensity failure */
	else if( strcmp( chMode, "eden" ) == 0 )
	{
		/* record number of electron density failures */
		++conv.nNeFail;

		if( called.lgTalk )
		{
			fprintf( ioQQQ, 
				" convfail: electron density fails, correct=%11.3e "
				" assumed=%11.3e zone=%3ld.", 
			  phycon.EdenTrue, phycon.eden, nzone );

			/* some extra information that may be printed */
			/* heating cooling failure */
			if( !conv.lgConvTemp )
			{
				fprintf( ioQQQ, "  Temperature failure also." );
			}

			/* heating cooling failure */
			if( !conv.lgConvIoniz )
			{
				fprintf( ioQQQ, "  Ionization failure also." );
			}

			/* electron density is oscillating */
			if( conv.lgEdenOscl )
			{
				fprintf( ioQQQ, "  Electron density oscillating." );
			}

			/* heating cooling oscillating */
			if( conv.lgCmHOsc )
			{
				fprintf( ioQQQ, "  Heating-cooling oscillating." );
			}
		}
		fprintf( ioQQQ, " \n");
	}

	else if( strcmp( chMode, "ioni" ) == 0 )
	{
		/* ionization failure */
		++conv.nIonFail;
		if( called.lgTalk )
		{
			fprintf( ioQQQ, " convfail: ionization not converged %3ld Z=%4ld \n", 
			  conv.nIonFail, nzone );
		}
	}

	/* rest of routine is temperature failure */
	else if( strcmp( chMode, "temp" ) == 0 )
	{
		++conv.nTeFail;
		if( called.lgTalk )
		{
			fprintf( ioQQQ, 
				" convfail te fail %2ld Z=%4ld TE=%10.4e ERROR=%10.2e"
				" HTOT=%10.2e CTOT=%10.2e DERR=%10.2e DTe=%10.2e IonDone%1c\n", 
			  conv.nTeFail, nzone, phycon.te, (heat.htot - cooling.ctot)/
			  heat.htot, heat.htot, cooling.ctot, cooling.dCoolmHeatdT, thermal.dTemper, 
			  TorF(conv.lgConvIoniz) );

			if( conv.lgTOscl && conv.lgCmHOsc )
			{
				fprintf( ioQQQ, " Temperature and d(Cool-Heat)/dT were both oscillating.\n" );
			}

			else if( conv.lgTOscl )
			{
				fprintf( ioQQQ, " Temperature was oscillating.\n" );
			}

			else if( conv.lgCmHOsc )
			{
				fprintf( ioQQQ, " d(Cool-Heat)/dT was oscillating.\n" );
			}

			/* not really a temperature failure, but something else */
			if( !conv.lgConvIoniz )
			{
				fprintf( ioQQQ, " Solution not converged due to %10.10s\n", 
				  conv.chConvIoniz );
			}
		}
	}
	else
	{
		fprintf( ioQQQ, " ConvFail called with insane mode %s \n", 
		  chMode );
		ShowMe();
		puts( "[Stop in ConvFail]" );
		cdEXIT(1);
	}

	/* increment total number of failures */
	++conv.nTotalFailures;

	/* now see how many total failures we have, and if it is time to abort */
	/* remember which zone this is */
	conv.ifailz[MIN2(conv.nTotalFailures,10)-1] = nzone;

	/* remember the relative error
	 * convert to single precision for following max, abs (vax failed here) */
	relerror = fabs((heat.htot-cooling.ctot)/ heat.htot);

	conv.failmx = MAX2(conv.failmx,(float)relerror);
	/*conv.failmx = (float)fmax(conv.failmx,fabs((float)((heat.htot-cooling.ctot)/
	  heat.htot)));*/

	if( conv.nTotalFailures < conv.LimFail )
	{ 
#		ifdef DEBUG_FUN
		fputs( " <->ConvFail()\n", debug_fp );
#		endif
		return;
	}

	/* punt */
	if( called.lgTalk )
	{
		fprintf( ioQQQ, " Stop after%3ld convergence failures.\n", 
		  conv.LimFail );
		fprintf( ioQQQ, " This limit can be reset with the FAILURES command.\n" );

		/* only do map if requested */
		/* adjust range of punting map */
		MapPar.RangeMap[0] = (float)(phycon.te/100.);
		MapPar.RangeMap[1] = (float)MIN2(phycon.te*100.,9e9);

		if( conv.lgMap )
		{
			/* need to make printout out now, before disturbing soln with map */
			PrtZone();
			punt("punt");
		}
	}

	/* return out from here and let lgEndFun catch lgBusted set,
	 * and generate normal output there */
	called.lgBusted = TRUE;
	if( called.lgTalk )
	{
		fprintf( ioQQQ, " ConvFail sets lgBusted since nTeFail>=LimFail%5ld%5ld\n", 
		  conv.nTeFail, conv.LimFail );
	}

#	ifdef DEBUG_FUN
	fputs( " <->ConvFail()\n", debug_fp );
#	endif
	return;
}
