/* DynaEndIter called at end of iteration when advection is turned on */
/* DynaStartZone called at end of iteration when advection is turned on */
/* DynaIonize, called from ionize to evaluate advective terms for current conditions */
/* DynaChangeDensity, called from PressureChange to evaluate new density needed for
 * current conditions and wind solution, returns ratio of new to old density */
/* ParseDynamics parse the advection command */
/* 
 * >>chng 01 mar 16, incorporate advection within dynamical solutions
 * this file contains the routines that incorporeate effects of advection upon the
 * thermal and ionization solutions.  It was originally developed in March 2001 during
 * a visit to UNAM Morelia, in collaboration with Jane Arthur, Will Henney, and Robin Williams.
 * These routines generate the produces that the code needs elsewhere.  
 */
#include "cddefines.h"
#include "parse.h"
#include "struc.h"
#include "radius.h"
#include "hydrogenic.h"
#include "converge.h"
#include "nomole.h"
#include "physok.h"
#include "pressure.h"
#include "ionrange.h"
#include "ionfracs.h"
#include "phycon.h"
#include "radacl.h"
#include "wind.h"
#include "dynamics.h"
static int ipUpstream=0;

/* ============================================================================== */
/* DynaChangeDensity, called from PressureChange to evaluate new density needed for
 * current conditions and wind solution, returns ratio of new to old density,
 * called when wind velocity is negative for D-critical solutions  */
double DynaChangeDensity(void)
{
	double factor;
	static double rsave = -1.;
	static double vold = -1.;

#	ifdef DEBUG_FUN
	fputs( "<+>DynaChangeDensity()\n", debug_fp );
#	endif
	double error , slope , DenNew , RamNeeded;
	/* this is logic for d-critical flow solution, developed with Will Henney*/

	/* following makes sure old wind v only updated once per zone,
	 * difference between old and new velocity is used to control zoning in nextdr */
	if( radius.depth != rsave )
	{
		rsave = radius.depth;
		/* evaluate radiative acceleration */
		radacl();

		/* remember the current wind velocity */
		vold = wind.windv;
	}

	/* this is the desired current ram pressure */
	/*>>chng 01 apr 02, correct sign error in PresInteg caught by Robin Williams */
	RamNeeded = pressure.PressureInit + pressure.RamPresInit -
	  pressure.GasPres + pressure.PresInteg;
	  /*pressure.GasPres - pressure.PresInteg;*/

	/* this is the current ram pressure, at this location */
	pressure.RamPres = phycon.xMassDensity*
		POW2(wind.windv0*struc.DenMass[0]/phycon.xMassDensity);

	/* now form the difference between current and desired pressure */
	/* this is error, the amount we need to make close to zero */
	error = RamNeeded - pressure.RamPres;

	/* this is the derivative of the pressure, 
	 * negative if ram pressure dominated, positive if gas pressure dominates */
	slope = (pressure.GasPres - pressure.RamPres)/phycon.hden;

	/* this is updated density */
	DenNew = phycon.hden + error / slope;

	/* this is the correction factor, to multipy densities of species, to make error go to zero, 
	 * it can be quite large and we want only small changes to keep things linear*/
	factor = DenNew / phycon.hden;
	factor = MAX2( 0.99 , factor );
	factor = MIN2( 1.01 , factor );

	/* remember the current total pressure */
	/* >>chng 01 apr 02, remove PresInteg from this term, caught by Robin Williams */
	/*pressure.pnow = (float)pressure.RamPres + pressure.GasPres + pressure.PresInteg;*/
	pressure.pnow = (float)pressure.RamPres + pressure.GasPres;

	/* update the current desired pressure, 
	 * the differences between pnow and pcorec will be
	 * used in ConvPresTempEdenIoniz to determine whether the 
	 * current zone has converged  */
	/* >>chng 01 apr 02, add PresInteg term here, caught by Robin Williams */
	/*pressure.pcorec = pressure.PressureInit + pressure.RamPresInit;*/
	pressure.pcorec = pressure.PressureInit + pressure.RamPresInit + pressure.PresInteg;

	/* update the wind velocity from mass conservation */
	wind.windv = wind.windv0*struc.DenMass[0]/phycon.xMassDensity;

	fprintf(ioQQQ,
		"nz %li now %.2e corr %.2e error %.2e slope %.2e pgas %.2e pram %.2e pint %.2e hden %.2e \n",
		nzone,
		pressure.pnow , pressure.pcorec , error,slope,
		pressure.GasPres , pressure.RamPres, pressure.PresInteg, phycon.hden );

	fprintf(ioQQQ,
		"winv %.2e cad  %.2e \n",
		wind.windv,sqrt(5.*pressure.GasPres/3./phycon.xMassDensity));


	/* this happens if no possible solun, */
	/*if( RamNeeded <=0. || 
		fabs(1.-pressure.GasPres/pressure.RamPres) < 0.05 )*/
	if( RamNeeded <=0. )
	{
		/* this will say to stop */
		wind.lgVelPos = FALSE;
		/* do not change the density */
		factor = 1.;
		/* trick the routine above us to think we are converged */
		pressure.pnow = pressure.pcorec;
	}

	/* following used to control new DRAD */
	wind.dVeldRad = (float)(fabs(wind.windv-vold)/radius.drad);

	fprintf(ioQQQ,
		"vold %.2e vnew %.2e dVdr %.2e \n", vold , wind.windv ,wind.dVeldRad);

#	ifdef DEBUG_FUN
	fputs( " <->DynaChangeDensity()\n", debug_fp );
#	endif
	return factor ;
}

/* ============================================================================== */
/* DynaIonize, called from ionize to evaluate advective terms for current conditions,
 * calculates terms to add to ionization balance equation */
void DynaIonize(void)
{
	long int nelem, ion;

#	ifdef DEBUG_FUN
	fputs( "<+>DynaIonize()\n", debug_fp );
#	endif

	if( iteration > 1 )
	{
		double work = 	
		phycon.EnergyIonization/phycon.xMassDensity +
		0.5*POW2(dynamics.Velocity)+
		/* thermal plus PdV work term */
		5./2.*pressure.GasPres/phycon.xMassDensity;

		/* net cooling due to advection */
		dynamics.CoolHeat = phycon.xMassDensity/dynamics.timestep*
			(work - dynamics.work);
		/*fprintf(ioQQQ,"dynamics cool-heat\t%.3e\t%.3e\t%.3e\t%.3e\n",
			phycon.te, dynamics.CoolHeat,work, dynamics.work );*/

		/* temp deriv of cooling minus heating */
		dynamics.dCoolHeatdT = phycon.xMassDensity/dynamics.timestep*5./2.*
			pressure.GasPres/phycon.xMassDensity/phycon.te;

		/* second or greater iteration, have advective terms */
		/* this will evaluate advective terms for current physical conditions */
		assert( dynamics.HAtom > 0.);

		/* the rate (s^-1) atoms drift in from upstream, a source term for the ground */
		dynamics.HRecom = 
			dynamics.HAtom / dynamics.timestep/xIonFracs[2][ipHYDROGEN] ;

		dynamics.Photo = 1./ dynamics.timestep;

		for( nelem=ipHYDROGEN; nelem<=ipZINC; ++nelem)
		{
			for( ion=IonRange.IonLow[nelem]; ion<IonRange.IonHigh[nelem]; ++ion )
			{
				dynamics.Recomb[ion][nelem] = 
					/* the +1 in xIonFracs is due to offset from 0, which has abundance */
					dynamics.UpstreamIon[ion+1][nelem]/ dynamics.timestep/xIonFracs[ion+2][nelem] *
					phycon.xMassDensity/dynamics.pdenstr;
			}
		}
		assert( fabs(dynamics.HRecom - dynamics.Recomb[0][ipHYDROGEN])/dynamics.HRecom < 0.001 );
	}
	else
	{
		/* first iteration, return zero */
		dynamics.HRecom = 0. ;
		dynamics.CoolHeat = 0.;
		dynamics.dCoolHeatdT = 0.;
		dynamics.Photo = 0.;

		for( nelem=ipHYDROGEN; nelem<=ipZINC; ++nelem)
		{
			for( ion=0; ion<=nelem+2 ; ++ion )
			{
				dynamics.Recomb[ion][nelem] = 0.;
			}
		}
	}

#	ifdef DEBUG_FUN
	fputs( " <->DynaIonize()\n", debug_fp );
#	endif
	return;
}

/* ============================================================================== */
/* DynaStartZone called at start of zone calculation when advection is turned on */
void DynaStartZone(void)
{
	/* this routine is called at the start of a zone calculation, by StartZone,
	 *
	 * it sets deduced variables to zero if first iteration,
	 *
	 * if upstream depth is is outside the computed structure on previous iteration,
	 * return value at shielded face */

	/* this is index of upstream cell in structure stored from previous iteration */
	double upstream;
	long int nelem , ion;

#	ifdef DEBUG_FUN
	fputs( "<+>DynaStartZone()\n", debug_fp );
#	endif

	/* radius.depth is distance from illuminated face of cloud to outer edge of
	 * current zone, which has thickness radius.drad */

	/* find where the down stream point is, in previous iteration,
	 * NB neg sign since velocity is negative, we are looking for point
	 * where current gas cell was, in previous iteration */
	upstream = radius.depth - dynamics.Velocity*dynamics.timestep;

	/* do nothing if first iteration, or is upstream point is beyond shielded face */
	if( (iteration <2  )  )
	{
		for( nelem=ipHYDROGEN; nelem<=ipZINC; ++nelem)
		{
			for( ion=0; ion<=nelem+2 ; ++ion )
			{
				dynamics.UpstreamIon[ion][nelem] = 0.;
			}
		}
		ipUpstream = 0;
		dynamics.work = 0.;

#		ifdef DEBUG_FUN
		fputs( " <->DynaStartZone()\n", debug_fp );
#		endif
		return;
	}

	if( upstream >= struc.Old_depth[struc.nOld_zone-1] )
	{
		/* overrun scale */
		dynamics.HAtom = struc.Old_histr[struc.nOld_zone-1];
		for( nelem=ipHYDROGEN; nelem<=ipZINC; ++nelem)
		{
			for( ion=0; ion<=nelem+2 ; ++ion )
			{
				dynamics.UpstreamIon[ion][nelem] = struc.Old_xIonFracs[struc.nOld_zone-1][ion][nelem];
			}
		}
		dynamics.work = struc.Old_work[struc.nOld_zone-1];
		/* upstream density, grams per unit vol, to account for divergence of gas */
		dynamics.pdenstr = struc.Old_pdenstr[struc.nOld_zone-1];
	}
	else
	{
		/* now locate upstream point in previous stored structure,
		 * will be at or higher than we found previously */
		while( (struc.Old_depth[ipUpstream+1] < upstream ) && 
		   ( ipUpstream < struc.nOld_zone-2 ) )
		{
			++ipUpstream;
		}
		assert( ipUpstream <= struc.nOld_zone-2 );

		/* above loop will turn ipUpstream == nOld_zone-1 it computed structure has been overrun */
		dynamics.HAtom = struc.Old_histr[ipUpstream] +
			(struc.Old_histr[ipUpstream+1] - struc.Old_histr[ipUpstream])/
			(struc.Old_depth[ipUpstream+1] - struc.Old_depth[ipUpstream])*
			( upstream - struc.Old_depth[ipUpstream]);

		/* the advected work */
		dynamics.work = struc.Old_work[ipUpstream] +
			(struc.Old_work[ipUpstream+1] - struc.Old_work[ipUpstream])/
			(struc.Old_depth[ipUpstream+1] - struc.Old_depth[ipUpstream])*
			( upstream - struc.Old_depth[ipUpstream]);

		/* the upstream particles per unit vol */
		dynamics.pdenstr = struc.Old_pdenstr[ipUpstream] +
			(struc.Old_pdenstr[ipUpstream+1] - struc.Old_pdenstr[ipUpstream])/
			(struc.Old_depth[ipUpstream+1] - struc.Old_depth[ipUpstream])*
			( upstream - struc.Old_depth[ipUpstream]);

		for( nelem=ipHYDROGEN; nelem<=ipZINC; ++nelem)
		{
			for( ion=0; ion<=nelem+2 ; ++ion )
			{
				dynamics.UpstreamIon[ion][nelem] = 
					struc.Old_xIonFracs[ipUpstream][ion][nelem] +
			(struc.Old_xIonFracs[ipUpstream+1][ion][nelem] - struc.Old_xIonFracs[ipUpstream][ion][nelem])/
			(struc.Old_depth[ipUpstream+1] - struc.Old_depth[ipUpstream])*
			( upstream - struc.Old_depth[ipUpstream]);
			}
		}
		assert( fabs(dynamics.HAtom - dynamics.UpstreamIon[1][ipHYDROGEN])/dynamics.HAtom < 0.001 );
	}

#	ifdef DEBUG_FUN
	fputs( " <->DynaStartZone()\n", debug_fp );
#	endif
	return;
}

/* ============================================================================== */
/* DynaEndIter called at end of iteration when advection is turned on */
void DynaEndIter(void)
{
	/* this routine is called by RestartIter at the end of an iteration 
	 * when advection is turned on.  currently it only derives a 
	 * timestep by looking at the spatial derivative of
	 * some stored quantities */

	/* largest desired fractional change in quantities */
	const double FRAC = 0.1;
	long i;
	double BigDeriv,
		dr,
		rec_time;

#	ifdef DEBUG_FUN
	fputs( "<+>DynaEndIter()\n", debug_fp );
#	endif

	/* reset the upstream counter */
	ipUpstream = 0;

	dr = 5e15;

	/* look for largest change in protron density, wrt depth, to
	 * get limit to timestep */
	BigDeriv = 0.;
	for( i=0; i<struc.nOld_zone-1; ++i )
	{
		BigDeriv = MAX2( BigDeriv , 
			fabs(struc.Old_hiistr[i+1]-struc.Old_hiistr[i])/struc.Old_hiistr[i]/
			(struc.Old_depth[i+1]-struc.Old_depth[i]) );
	}

	/* the phyiscal thickness (cm) over which quantity varied by this amount */
	if( BigDeriv < SMALLFLOAT )
	{
		dr = 0.;
	}
	else
	{
		dr = FRAC / BigDeriv;
	}
	/* the timestep corresponding to this dr and velocity */
	dynamics.timestep = dr / -dynamics.Velocity;

	/* the timestep corresponding to this dr and velocity */
	dynamics.timestep = dr / -dynamics.Velocity;

	/* recombination timescale for first zone */
	rec_time = 1./(struc.ednstr[0]*2.90e-10*pow(struc.testr[0],-0.77));

	/* hack in rec timescale for now */
	dynamics.timestep = rec_time;
	dr = dynamics.timestep * -dynamics.Velocity ;

	assert( dynamics.timestep > 0. );

	fprintf(ioQQQ," DynaEndIter, BigDeriv=%.2e dr=%.2e timestep=%.2e\n",
		BigDeriv , dr , dynamics.timestep );

#	ifdef DEBUG_FUN
	fputs( " <->DynaEndIter()\n", debug_fp );
#	endif
	return;
}

/* ============================================================================== */
/*ParseDynamics parse the dynamics command */
void ParseDynamics(char *chCard )
{
	long int i;
	int lgEOL;

#	ifdef DEBUG_FUN
	fputs( "<+>ParseDynamics()\n", debug_fp );
#	endif

	/* turn on advection */
	dynamics.lgAdvection = TRUE;

	/* turn off prediction of next zone's temperature, as guessed in StartZone */
	phycon.lgPredNextTe = FALSE;

	/* turn off both CO and H2 networks since advection not included, also say physicsl
	 * conditions are not ok*/
	nomole.lgNoH2Mole = TRUE;
	nomole.lgNoCOMole = TRUE;
	physok.lgPhysOK = FALSE;

	/* use the new temperature solver */
	strcpy( conv.chSolverEden , "new" );

	/* increse precision of solution */
	phycon.EdenError = 1e-3f;

	/* scan off the velocity, entered in km/s, converted to cm/s */
	i = 5;
	dynamics.Velocity = FFmtRead(chCard,&i, LINELENGTH,&lgEOL)*1e5;
	if( lgEOL )
	{
		NoNumb(chCard);
	}

	/* make it negative, must be for an ionizaiton front */
	if( dynamics.Velocity > 0. )
		dynamics.Velocity *= -1.;

#	ifdef DEBUG_FUN
	fputs( " <->ParseDynamics()\n", debug_fp );
#	endif
	return;
}
