/*HydroPesc evaluate escape and destruction probabilities for hydrogen lines,
 * called by RTMake */
#include "cddefines.h"
#include "physconst.h"
#include "taulines.h"
#include "hydrogenic.h"
#include "strk.h"
#include "phycon.h"
#include "converge.h"
#include "opacity.h"
#include "twophoton.h"
#include "trace.h"
#include "hlife.h"
#include "rfield.h"
#include "ionfracs.h"
#include "wind.h"
#include "bowoi.h"
#include "p8446.h"
#include "rt.h"
#include "rtescprob.h"
/* needed in excluded debug print statement */
#include "fe2ovr.h"

void HydroPesc(
	/* ipZ is on C scale, 0 for H, 1 for He, etc */
	long int ipZ ,
	/* flag saying whether we should do escape prob  and reevaluate A and opac (TRUE) or
	 * only dest probs (FALSE).  This is called with TRUE one time per zone,
	 * when optical depths are updated.  It is called FALSE many times, within ionzie,
	 * each time the opacity changes */
	int lgDoEsc )
{
	static int lgTOnSave;
	long int ipHi, 
	  ipLo, 
	  limit;
	double abund=0., 
	  coloi, 
	  es, 
	  factor, 
	  tout,
	  z4;/* physical scale z to the 4th power, used often below */
	float dest=0.f, 
	  esin;
	float FracNew ;
	/* this will be used to save old vals of destruction prob in case
	 * we need to use mean of old and new, as is necessary when ots oscillations
	 * take place */
	float DespSave[NHYDRO_MAX_LEVEL];
	float ri2s1s;
	float ri1s2s;

#	ifdef DEBUG_FUN
	fputs( "<+>HydroPesc()\n", debug_fp );
#	endif

	/* check that we were called with valid charge */
	assert( ipZ >= 0);
	assert( ipZ < LIMELM );

	/* will need this for some scaling laws - physical Z to the 4th power*/
	z4 = POW2(ipZ+1.);
	z4 *= z4;

	if( trace.lgTrace )
	{
		fprintf( ioQQQ, 
			"          HydroPesc ipZ=%ld called\n", 
		  ipZ );
	}

	if( lgDoEsc )
	{
		/* hydrogen lines opacity is function of A's
		 * above 15 just use actual A's for high density limit
		 * only change opacities for levels between 4 and 15 */
		/* NB! must never set opacity for 2s-1s two-photon transition here, since
		 * the "line" can become optically thick if treated as a line.  It is a 
		 * continuum and the opacity is not related to the A by the usual expression */
		limit = MIN2(15,nHLevels[ipZ]);
		/* do Paschen and higher lines first, 
		 * do balmer below since must separate 2s and 2p */
		for( ipHi=4; ipHi <= limit; ipHi++ )
		{
			for( ipLo=3; ipLo < ipHi; ipLo++ )
			{
				HydroLines[ipZ][ipHi][ipLo].Aul = 
					(float)(hlife.HyLife[ipHi]*
					HydroBranch(ipHi,ipLo,ipZ+1)*z4);
				assert(HydroLines[ipZ][ipHi][ipLo].Aul > 0.);

				/* make self-consistent opacity, convert new As back into opacities */
				HydroLines[ipZ][ipHi][ipLo].opacity = 
					(float)(HydroLines[ipZ][ipHi][ipLo].Aul*
				  2.2448e-26*HydroLines[ipZ][ipHi][ipLo].gHi/
				  HydroLines[ipZ][ipHi][ipLo].gLo*
				  POW3(RYDLAM/HydroLines[ipZ][ipHi][ipLo].EnergyRyd));

				/* check that results are ok */
				assert(HydroLines[ipZ][ipHi][ipLo].opacity > 0.);
			}
		}

		/* the actual branching ratio from high levels down to
		 * 2s and 2p depend on the density.  the code goes to
		 * the correct low and high density limits - I know of
		 * nothing better than can be done. */
		factor = MAX2( 0.25 , 0.32 - 0.07*phycon.eden/(phycon.eden + 1e7) );

		/* treat 2s = to 2p for HydroBranch, which returns total to 2s+2p */
		for( ipHi=4; ipHi <= limit; ipHi++ )
		{
			/* get new effective A for this density and temperature */
			HydroLines[ipZ][ipHi][ipH2s].Aul = 
				(float)(hlife.HyLife[ipHi]*
				factor*HydroBranch(ipHi,2,ipZ+1)*z4);

				/* check that results are ok */
			assert(HydroLines[ipZ][ipHi][ipH2s].Aul > 0.);

			/* do 2p by scaling relative to 2s */
			HydroLines[ipZ][ipHi][ipH2p].Aul = (float)(
				HydroLines[ipZ][ipHi][ipH2s].Aul / factor *( 1. - factor )); 

				/* check that results are ok */
			assert(HydroLines[ipZ][ipHi][ipH2p].Aul > 0.);

			/* make self-consistent opaciity for 2s, from A */
			HydroLines[ipZ][ipHi][ipH2s].opacity = 
				(float)(HydroLines[ipZ][ipHi][ipH2s].Aul*
			  2.2448e-26*HydroLines[ipZ][ipHi][ipH2s].gHi/
			  HydroLines[ipZ][ipHi][ipH2s].gLo*
			  POW3(RYDLAM/HydroLines[ipZ][ipHi][ipH2s].EnergyRyd));

				/* check that results are ok */
			assert(HydroLines[ipZ][ipHi][ipH2s].opacity > 0.);

			/* make self-consistent opaciity for 2p, from A */
			HydroLines[ipZ][ipHi][ipH2p].opacity = 
				(float)(HydroLines[ipZ][ipHi][ipH2p].Aul*
			  2.2448e-26*HydroLines[ipZ][ipHi][ipH2p].gHi/
			  HydroLines[ipZ][ipHi][ipH2p].gLo*
			  POW3(RYDLAM/ HydroLines[ipZ][ipHi][ipH2p].EnergyRyd));

				/* check that results are ok */
			assert(HydroLines[ipZ][ipHi][ipH2p].opacity > 0.);
		}
	}
	/* end test branch lgDoEsc */

	/* now update escape and destruction prob */

	if( xIonFracs[ipZ+2][ipZ] > 1e-30 )
	{
		factor = xIonFracs[ipZ+2][ipZ];
	}
	else
	{
		/* case where almost no parent ion - this will make
		 * very large line opacity, so background dest small */
		factor = 1.;
	}

	/* save destruction probs for Lyman lines in case ots rates oscillate */
	for( ipHi=ipH2p; ipHi<= nHLevels[ipZ]; ++ipHi )
	{
		DespSave[ipHi] = HydroLines[ipZ][ipHi][ipH1s].Pdest;
	}

	/* first will be static solution */
	if( wind.windv == 0. )
	{
		/* hydrogenic lyman lines special since outward optical depths always set,
		 * trick routines for Lyman lines only */
		lgTOnSave = opac.lgTauOutOn;
		opac.lgTauOutOn = TRUE;

		/* first do lyman alpha lya la, but only if optical depths not overrun */
		/* >>chng 01 apr 01, from 0.9 to 0.99 since outer edge better defined */
		tout = HydroLines[ipZ][ipH2p][ipH1s].TauTot*0.99 - 
			HydroLines[ipZ][ipH2p][ipH1s].TauIn;

		/* must temporarily make ipLnPopOpc physical */
		HydroLines[ipZ][ipH2p][ipH1s].PopOpc *= 
			(float)factor;

		/* generate escape prob, pumping rate, destruction prob, 
		 * inward outward fracs  */
		RTMakeStat(&HydroLines[ipZ][ipH2p][ipH1s] , lgDoEsc );

		/* go back to original units so that final correction ok */
		HydroLines[ipZ][ipH2p][ipH1s].PopOpc /= 
			(float)factor;

		/* >>>chng 99 dec 18, repair dest prob that got clobbered in call to
		 * RTMakeStat, since will not be evaluated when tout not positive */
		HydroLines[ipZ][ipH2p][ipH1s].Pdest = DespSave[ipH2p];

		/* only update La dest prob is good outward optical depths */
		if( tout > 0. )
		{
			tout = HydroLines[ipZ][ipH2p][ipH1s].TauTot - 
			  HydroLines[ipZ][ipH2p][ipH1s].TauIn;

			abund = HydroLines[ipZ][ipH2p][ipH1s].PopOpc*xIonFracs[ipZ+2][ipZ];

			/* the descruction prob comes back as dest */
			HydroLines[ipZ][ipH2p][ipH1s].Pesc = 
				(float)(escla(&esin, &dest,abund,ipZ));

			/* this is current destruction rate */
			HydroLines[ipZ][ipH2p][ipH1s].Pdest = dest;

			/* >>>chng 99 apr 16, average of old and new rates to damp opacity - Lya ots
			 * rates when opacity is large, as in HeI trip opac in Balmer continuum.
			 * the pphheonly.in test case exposes this problem */
			{
				/*@-redef@*/
				enum {DEBUG=FALSE};
				/*@+redef@*/
				if( DEBUG && ipZ==0 && nzone > 268 && iteration == 2)
				{
					fprintf(ioQQQ,
						"z%3li Lya eval Pdest popl\t%g\tconopac\t%g\tPdest\t%g\t tot in\t%g\t%g\n",
						nzone ,
						abund , 
						opac.opac[HydroLines[ipZ][ipH2p][ipH1s].ipCont-1],
						HydroLines[ipZ][ipH2p][ipH1s].Pdest, 
						HydroLines[ipZ][ipH2p][ipH1s].TauTot,
						HydroLines[ipZ][ipH2p][ipH1s].TauIn);
				}
			}

			/*  this is fraction of line which is inward escaping */
			HydroLines[ipZ][ipH2p][ipH1s].FracInwd = rt.fracin;
			/*if(ipZ==1)
				fprintf(ioQQQ," desp=%g\n", dest );*/
		}

		/* now do remainder of Lyman lines, skipping 2s */
		ipLo=ipH1s;
		for( ipHi=3; ipHi <= nHLevels[ipZ]; ipHi++ )
		{
			/* must temporarily make ipLnPopOpc physical */
			HydroLines[ipZ][ipHi][ipLo].PopOpc *= 
				(float)factor;

			/* generate escape prob, pumping rate, destruction prob, 
			 * inward outward fracs  */
			RTMakeStat(&HydroLines[ipZ][ipHi][ipLo] , lgDoEsc );

			/* go back to original units so that final correction ok */
			HydroLines[ipZ][ipHi][ipLo].PopOpc /= 
				(float)factor;
		}

		/* reset the flag, so only Lyman lines forced to include outward */
		opac.lgTauOutOn = lgTOnSave;

		/* this loop for Balmer lines which are special,
		 * because must bring 2s and 2p together */
		/* now do 2s and 2p, must bring optical depths together */
		for( ipHi=3; ipHi <= nHLevels[ipZ]; ipHi++ )
		{
			/* these are used for saving current state of two lines */
			float tauin2s , tauin2p , tauout2s , tauout2p , opac2s , opac2p;

			/* 2s inward and total optical depths */
			tauin2s =  HydroLines[ipZ][ipHi][ipH2s].TauIn;
			tauout2s = HydroLines[ipZ][ipHi][ipH2s].TauTot;

			/* 2p inward and total optical depths */
			tauin2p =  HydroLines[ipZ][ipHi][ipH2p].TauIn;
			tauout2p = HydroLines[ipZ][ipHi][ipH2p].TauTot;

			/* add inward optical depths together */
			HydroLines[ipZ][ipHi][ipH2s].TauIn += 
			HydroLines[ipZ][ipHi][ipH2p].TauIn;

			HydroLines[ipZ][ipHi][ipH2p].TauIn = 
			HydroLines[ipZ][ipHi][ipH2s].TauIn;

			/* add both total optical depths together */
			HydroLines[ipZ][ipHi][ipH2s].TauTot += 
			HydroLines[ipZ][ipHi][ipH2p].TauTot;

			HydroLines[ipZ][ipHi][ipH2p].TauTot = 
			HydroLines[ipZ][ipHi][ipH2s].TauTot;

			/* current 2s and 2p opacities */
			opac2s = HydroLines[ipZ][ipHi][ipH2s].PopOpc;
			opac2p = HydroLines[ipZ][ipHi][ipH2p].PopOpc;

			/* add opacities together */
			HydroLines[ipZ][ipHi][ipH2s].PopOpc += 
			HydroLines[ipZ][ipHi][ipH2p].PopOpc;

			HydroLines[ipZ][ipHi][ipH2p].PopOpc = 
			HydroLines[ipZ][ipHi][ipH2s].PopOpc;

			/* must temporarily make ipLnPopOpc physical */
			HydroLines[ipZ][ipHi][ipH2s].PopOpc *= 
				(float)factor;
			HydroLines[ipZ][ipHi][ipH2p].PopOpc *= 
				(float)factor;

			/* generate escape prob, pumping rate, destruction prob, 
			 * inward outward fracs  */
			RTMakeStat(&HydroLines[ipZ][ipHi][ipH2s] , lgDoEsc );
			RTMakeStat(&HydroLines[ipZ][ipHi][ipH2p] , lgDoEsc );

			HydroLines[ipZ][ipHi][ipH2s].PopOpc =   opac2s;
			HydroLines[ipZ][ipHi][ipH2s].TauIn =   tauin2s;
			HydroLines[ipZ][ipHi][ipH2s].TauTot = tauout2s;

			/* 2p inward and total optical depths */
			HydroLines[ipZ][ipHi][ipH2p].PopOpc =   opac2p;
			HydroLines[ipZ][ipHi][ipH2p].TauIn =   tauin2p;
			HydroLines[ipZ][ipHi][ipH2p].TauTot = tauout2p;
		}

		/* now do Paschen and higher lines */
		for( ipLo=3; ipLo <= (nHLevels[ipZ] - 1); ipLo++ )
		{
			for( ipHi=ipLo+1; ipHi <= nHLevels[ipZ]; ipHi++ )
			{
				/* must temporarily make ipLnPopOpc physical */
				HydroLines[ipZ][ipHi][ipLo].PopOpc *= 
					(float)factor;

				/* generate escape prob, pumping rate, destruction prob, 
				 * inward outward fracs  */
				RTMakeStat(&HydroLines[ipZ][ipHi][ipLo] , lgDoEsc );

				/* go back to original units so that final correction ok */
				HydroLines[ipZ][ipHi][ipLo].PopOpc /= 
					(float)factor;
			}
		}

		/* 666    C90 had the following code and warning about oscillations
		 *  highest level is a mess, set escape prob to one, actually a blend
		 *  if this is removed, then oscillations in totally optically thick
		 *  blr occur, for cases where all lines are optically thick
		 *  see model badbugs/bug0.in */
		HydroLines[ipZ][nHLevels[ipZ]][ipH1s].Pesc = (float)(
			MAX2( HydroLines[ipZ][nHLevels[ipZ]][ipH1s].Pesc , 0.1));

		for( ipLo=ipH2s; ipLo <= (nHLevels[ipZ] - 1); ipLo++ )
		{
			HydroLines[ipZ][nHLevels[ipZ]][ipLo].Pesc = 1.;
		}
	}
	else
	{
		/* this is wind solution branch */
		/* hydrogenic lyman lines special since outward optical depths always set,
		 * trick routines for Lyman lines only */
		lgTOnSave = opac.lgTauOutOn;
		opac.lgTauOutOn = TRUE;

		/* say that no outward optical depths, but we are still ok */
		tout = 0.;

		/* windy model */
		for( ipLo=ipH1s; ipLo <= (nHLevels[ipZ] - 1); ipLo++ )
		{
			for( ipHi=MAX2((long)ipH2p,ipLo+1); ipHi <= nHLevels[ipZ]; ipHi++ )
			{
				/* must temporarily make ipLnPopOpc physical */
				HydroLines[ipZ][ipHi][ipLo].PopOpc *= 
					(float)factor;

				RTMakeWind(&HydroLines[ipZ][ipHi][ipLo] , lgDoEsc );

				/* go back to original units so that final correction ok */
				HydroLines[ipZ][ipHi][ipLo].PopOpc /= 
					(float)factor;
			}
			/* reset the flag, so only Lyman lines forced to include outward */
			opac.lgTauOutOn = lgTOnSave;
		}
	}

	/* this is option to damp out Lyman line dest probs if ots rates oscillating */
	if( conv.lgOscilOTS )
	{
		/* this is damper used to stop oscillations when now present*/
		FracNew = 0.2f;
	}
	else
	{
		/* this is damper used to stop oscillations even when not detected */
		FracNew = 0.5f;
	}
	{
		/* following should be set true to print ots contributors */
		/*@-redef@*/
		enum {DEBUG=FALSE};
		/*@+redef@*/
		if( DEBUG && ipZ==0 )
		{
			fprintf(ioQQQ,"Lya aver DespSave\t%g\tPdest\t%g\n",
				DespSave[ipH2p],
				HydroLines[ipZ][ipH2p][ipH1s].Pdest);
		}
	}
	/* >>chng 01 apr 01, add test for tout > 0, only in this case will
	 * new solution even be attempted */
	if( tout >= 0. )
	{
		for( ipHi=ipH2p; ipHi<= nHLevels[ipZ]; ++ipHi )
		{
			/*lint -e771 DespSave possibly not initialized */
			HydroLines[ipZ][ipHi][ipH1s].Pdest = (1.f-FracNew)*DespSave[ipHi] +
				FracNew * HydroLines[ipZ][ipHi][ipH1s].Pdest;
			/*lint +e771 DespSave possibly not initialized */
		}
	}

	{
		/* following should be set true to print ots contributors */
		/*@-redef@*/
		enum {DEBUG=FALSE};
		/*@+redef@*/
		if( DEBUG && ipZ==0 )
		{
			fprintf(ioQQQ,"Lya ave desp\t%g\t popl\t%g\tconopac\t%g\tbeta\t%g\tfracnew%g\n",
				dest , 
				abund , 
				DespSave[ipH2p],
				HydroLines[ipZ][ipH2p][ipH1s].Pdest, FracNew);
		}
	}

	/* reset the flag */
	opac.lgTauOutOn = lgTOnSave;

	/* find the rate of total and induced two photon */
	twopht(ipZ, &ri2s1s , &ri1s2s , ipHYDROGEN );

	HydroLines[ipZ][ipH2s][ipH1s].pump = ri1s2s;
	/* >>chng 01 jan 18, us derived two-photon, the sum of spontaneous and induced */
	HydroLines[ipZ][ipH2s][ipH1s].Aul = ri2s1s;

	/* hydrogen-only solutions */
	if( ipZ == 0 )
	{
		/* find Stark escape probabilities */
		strk();
		/* electron scattering escape */
		es = phycon.eden*6.65e-25;
		for( ipLo=ipH1s; ipLo <= (nHLevels[ipZ] - 1); ipLo++ )
		{
			for( ipHi=MAX2((long)ipH2p,ipLo+1); ipHi <= nHLevels[ipZ]; ipHi++ )
			{
				hydro.esesc[ipZ][ipLo][ipHi] = (float)(MIN2(1.,5.9*es/(es+
				  HydroLines[ipZ][ipHi][ipLo].PopOpc*
				  HydroLines[ipZ][ipHi][ipLo].opacity))*
				  xIonFracs[ipZ+2][ipZ]*
				  MAX2(0.,1.-HydroLines[ipZ][ipHi][ipLo].Pesc-
				  HydroLines[ipZ][ipHi][ipLo].Pdest));
				hydro.esesc[ipZ][ipLo][ipHi] = 0.;
			}
		}
		hydro.esesc[ipZ][ipH1s][ipH2s] = 0.;
		/*fixit(); in above and in loop, electrons cattering escape set to 0 since got
		           * very large in case b test */

		/* do the 8446 problem */
		p8446(&coloi);
		HydroLines[0][3][ipH1s].Pesc = bowoi.pmph31/
		  HydroLines[0][3][ipH1s].Aul;

		if( trace.lgTrace && trace.lgHBugFull )
		{
			fprintf( ioQQQ, "       HydroPesc calls P8446 who found pmph31=%10.2e\n", 
			  bowoi.pmph31 );
		}

		/* add on destruction of hydrogen Lya by FeII
		 * now add in FeII deexcitation via overlap,
		 * but only as loss, not photoionization, source
		 * dstfe2lya is Ly-alpha deexcit by FeII overlap - conversion into FeII em */
		/* find FeII overlap destruction rate, 
		 * this does NOTHING when large FeII atom is turned on */
		fe2ovr();

		/* this was introduced in the fort - c conversion, and is a bug since
		 * dest feii added to ots dest */
		/*HydroLines[0][ipH2p][ipH1s].Pdest = MIN2(1.f,
		  HydroLines[0][ipH2p][ipH1s].Pdest + hydro.dstfe2lya);*/
		/* >>chng 00 jan 06, let dest be large than 1 to desaturate the atom */
		/* >>chng 01 apr 01, add test for tout >= 0., 
		 * must not add to Pdest when it has not been refreshed here */
		if( tout >= 0. )
			HydroLines[0][ipH2p][ipH1s].Pdest += hydro.dstfe2lya;

		{
			/*@-redef@*/
			enum {DEBUG=FALSE};
			/*@+redef@*/
			if( DEBUG )
			{
				fprintf(ioQQQ,"hydropescdebugg\t%li\t%.2e\t%.2e\t%.2e\t%.2e\t%.2e\t%.2e\n", 
					nzone, 
					HydroLines[0][2][0].pump,
					HydroLines[0][4][3].TauIn,
					HydroLines[0][4][3].TauCon,
					HydroLines[0][4][3].pump,
					rfield.flux[HydroLines[0][4][3].ipCont-1],
					rfield.OccNumbIncidCont[HydroLines[0][4][3].ipCont-1]
					);
			}
		}
	}

	{
		/*@-redef@*/
		enum {DEBUG=FALSE};
		/*@+redef@*/
		if( DEBUG )
		{
			if( ipZ==0&& nzone>164 )
			{
				fprintf(ioQQQ,"hydropescdebugg\t%li\t%.2e\t%.2e\t%.2e\t%.2e\t%.2e\n", 
					nzone, 
					HydroLines[0][2][0].TauIn,
					HydroLines[0][2][0].TauTot,
					HydroLines[0][2][0].Pdest,
					HydroLines[0][2][0].Pesc,
					tout);
			}
		}
	}

	/*fixit(); this was cut out of pesc and pasted here.  this needs to
				 * be added back into the total rates, even though it is not important */
#	if 0
	/* increase escape probability by scattering and stark */
	HydroLines[ipZ][ipHi][ipLo].Pesc = (float)(fmin(1.,
	  HydroLines[ipZ][ipHi][ipLo].Pesc+
	  hydro.pestrk[ipLo][ipHi]+hydro.esesc[ipZ][ipLo][ipHi]));
#	endif

#	ifdef DEBUG_FUN
	fputs( " <->HydroPesc()\n", debug_fp );
#	endif
	return;
}

