/*ParsePunch parse the punch command */
#include "cddefines.h"
#include "elementnames.h"
#include "pnunit.h"
#include "input.h"
#include "ipdr.h"
#include "iph2.h"
#include "levery.h"
#include "zonecnt.h"
#include "ioreco.h"
#include "opacity.h"
#include "mappar.h"
#include "pop371.h"
#include "date.h"
#include "punconv.h"
#include "punpoint.h"
#include "punrlttype.h"
#include "grainvar.h"
#include "punlin.h"
#include "physconst.h"
#include "parse.h"
/* check for keyword UNITS on line, then scan wavelength or energy units if present */
static void ChkUnits( char *chCard);


/* NB NB NB NB NB NB NB NB NB NB
 *
 * if any "special" punch commands are added (commands that copy the file pointer
 * into a separate variable, e.g. like PUNCH _DR_), be sure to add that file pointer
 * to InitPunchFiles and ClosePunchFiles !!!
 *
 * PUNCH FILE POINTERS SHOULD NEVER BE ALTERED BY ROUTINES OUTSIDE THIS MODULE !!!
 *
 * hence initializations of punch file pointers should never be included in zero() !!
 * the pointer might be lost without the file being closed
 * 
 * NB NB NB NB NB NB NB NB NB NB */


void ParsePunch(char *chCard )
{
	char chLabel[81];
	int lgEOL ,
	  lgHit ;
	/* pointer to column across line image for free format number reader*/
	long int ipFFmt, 
	  i,
	  j,
	  nelem;

#	ifdef DEBUG_FUN
	fputs( "<+>ParsePunch()\n", debug_fp );
#	endif

	/* while optimizing, punch commands are not executed -> return */
	if( !pnunit.lgOpenUnits )
	{
#		ifdef DEBUG_FUN
		fputs( " <->ParsePunch()\n", debug_fp );
#		endif
		return;
	}

	/* check that limit not exceeded */
	if( pnunit.npunch > LIMPUN )
	{
		fprintf( ioQQQ, 
			"The limit to the number of PUNCH options is%3ld.  Increase LIMPUN in pnunit.h if more are needed.\n", 
		  LIMPUN );
		puts( "[Stop in ParsePunch]" );
		cdEXIT(1);
	}

	/* LAST keyword is an option to punch only on last iteration */
	pnunit.lgPunLstIter[pnunit.npunch] = lgMatch("LAST",chCard);

	/* NOCLOBBER keyword is an option to append to a given file instead of overwriting it */
	pnunit.lgNoClobber[pnunit.npunch] = lgMatch("NOCL",chCard);

	/* 
	 * get file name for this punch output.
	 * GetQuote does the following -
	 * first copy original version of file name into chLabel, 
	 * string does include null termination.
	 * set filename in OrgCard and second parameter to spaces so 
	 * that not picked up below as keyword
	 */
	GetQuote( chLabel , chCard );

	/* check that name is not same as opacity.opc, a special file */
	if( strcmp(chLabel , "opacity.opc") == 0 )
	{
		fprintf( ioQQQ, " ParsePunch will not allow punch file name %s, please choose another.\n",
			chLabel);
		puts( "[Stop in ParsePunch]" );
		cdEXIT(1);
	}

	/* only open if file has not already been opened during a previous call (should never happen!) */
	if( pnunit.ipPnunit[pnunit.npunch] == NULL )
	{
		/* pnunit.npunch is set to 0 in ZERO, 
		 * pnunit.ipPnunit is the file pointer  */

		/* open the file with the name generated above */
		if( pnunit.lgNoClobber[pnunit.npunch] ) 
		{
			pnunit.ipPnunit[pnunit.npunch] = fopen( chLabel,"a+" );
		}
		else 
		{
			pnunit.ipPnunit[pnunit.npunch] = fopen( chLabel,"w" );
		}
		if( NULL == pnunit.ipPnunit[pnunit.npunch] )
		{
			fprintf( ioQQQ, " ParsePunch could not open punch file %s for writing.\n",
				chLabel);
			puts( "[Stop in ParsePunch]" );
			cdEXIT(1);
		}
	}

	/* usually results for each iteration are followed by a series of
	 * hash marks, ####, which fool excel.  This is an option to not print
	 * the line.  If the keyword NO HASH appears anywhere, on any punch command,
	 * the hash marks will never be produced */
	if( lgMatch("O HA",chCard) )
	{
		pnunit.lgHashEndIter = FALSE;
	}

	/* put version number and title of model on output file */
	fprintf( pnunit.ipPnunit[pnunit.npunch], 
		"# %s %s\n", 
	  date.chVersion, input.chTitle );

	/* this must come first since elements appear as sub-keywords */
	if( lgMatch("OPAC",chCard) )
	{

		/* check for keyword UNITS on line, then scan wavelength or energy units if present,
		 * units are copied into pnunit.chConPunEnr */
		ChkUnits(chCard);

		strcpy( pnunit.chPunch[pnunit.npunch], "OPAC" );
		if( lgMatch("TOTA",chCard) )
		{
			/* DoPunch will call popac to parse the subcommands
			 * punch total opacity */
			strcpy( opac.chOpcTyp, "TOTL" );
			fprintf( pnunit.ipPnunit[pnunit.npunch], 
				"#nu\tTot opac\tAbs opac\tScat opac\tAlbedo\telem\n" );
		}

		else if( lgMatch("FIGU",chCard) )
		{
			/* do figure for hazy */
			strcpy( opac.chOpcTyp, "FIGU" );
			fprintf( pnunit.ipPnunit[pnunit.npunch], 
				"#nu, H, He, tot opac\n" );
		}

		else if( lgMatch("GRAI",chCard) )
		{
			/* punch grain command, give optical properties of gains in calculation */
			strcpy( pnunit.chPunch[pnunit.npunch], "DUSO" );
			/* punch grain opacity command in twice, here and above in opacity
			 * chDst = 'OPAC' */
			fprintf( pnunit.ipPnunit[pnunit.npunch], 
				"#grain nu,  total,  absorpt,  scat\n" );
		}

		else if( lgMatch("SHEL",chCard) )
		{
			/* punch shells, a form of the punch opacity command for showing subshell crossections*/
			strcpy( pnunit.chPunch[pnunit.npunch], "OPAC" );

			/* punch subshell cross sections */
			strcpy( opac.chOpcTyp, "SHEL" );

			/* this is element */
			ipFFmt = 3;
			pnunit.punarg[0][pnunit.npunch] = (float)FFmtRead(chCard,&ipFFmt,
			  LINELENGTH,&lgEOL);

			/* this is ion */
			pnunit.punarg[1][pnunit.npunch] = (float)FFmtRead(chCard,&ipFFmt,
			  LINELENGTH,&lgEOL);

			/* this is shell */
			pnunit.punarg[2][pnunit.npunch] = (float)FFmtRead(chCard,&ipFFmt,
			  LINELENGTH,&lgEOL);

			if( lgEOL )
			{
				fprintf( ioQQQ, " there must be io unit, atom wght, ion, shell\n" );
				puts( "[Stop in ParsePunch]" );
				cdEXIT(1);
			}
			fprintf( pnunit.ipPnunit[pnunit.npunch], "#sub shell cross section\n" );
		}

		else
		{
			/* punch elenent opacity, produces n name.n files, one for each stage of 
			 * ionization.  the name is the 4-char version of the element's name, and
			 * n is the stage of ionization.  the file nameon the card is ignored.
			 * The code Stop in getpunchs after these files are produced. */

			/* this will be used as check that we did find an element on the command lines */
			lgHit = FALSE;
			for( nelem=0; nelem<LIMELM; ++nelem )
			{
				/* look for the name of one of the 30 elements */
				if( lgMatch(elementnames.chElementNameShort[nelem],chCard) )
				{
					strcpy( opac.chOpcTyp, elementnames.chElementNameShort[nelem] );
					lgHit = TRUE;
					break;
				}
			}

			/* lgHit was set true if an element was found, error if not */
			if( !lgHit )
			{
				fprintf( ioQQQ, " There must be a second keyword on the opacity command.  Sorry.\n" );
				fprintf( ioQQQ, " Options are total, figure, grain, or element name.\n" );
				puts( "[Stop in ParsePunch]" );
				cdEXIT(1);
			}
		}
	}

	else if( lgMatch("ABUN",chCard) )
	{
		/* punch abundances */
		strcpy( pnunit.chPunch[pnunit.npunch], "ABUN" );
		fprintf( pnunit.ipPnunit[pnunit.npunch], "#abundances\n" );
	}

	else if( lgMatch("ADVE",chCard) )
	{
		/* punch information relating to advection */
		strcpy( pnunit.chPunch[pnunit.npunch], "ADVE");
		fprintf( pnunit.ipPnunit[pnunit.npunch], "#avection\n" );
	}

	else if( lgMatch("AGES",chCard) )
	{
		/* punch ages */
		strcpy( pnunit.chPunch[pnunit.npunch], "AGES" );
		fprintf( pnunit.ipPnunit[pnunit.npunch], "#ages\n" );
	}

	else if( lgMatch("ASSE",chCard) )
	{
		/* punch asserts */
		strcpy( pnunit.chPunch[pnunit.npunch], "ASSE" );
		/* no need to print this standard line of explanation*/
		/*fprintf( pnunit.ipPnunit[pnunit.npunch], " asserts\n" );*/
	}

	/* punch charge transfer */
	else if( lgMatch("CHAR",chCard) )
	{
		/* NB in dopunch only the first three characters are compared to find this option,
		 * search for "CHA" */
		/* this is the AGN option, to produce a table for AGN */
		if( lgMatch(" AGN",chCard) )
		{
			strcpy( pnunit.chPunch[pnunit.npunch], "CHAG" );
			fprintf( pnunit.ipPnunit[pnunit.npunch], "#charge exchange rate coefficnt\n" );
		}
		else
		{
			/* punch charge transfer */
			strcpy( pnunit.chPunch[pnunit.npunch], "CHAR" );
			fprintf( pnunit.ipPnunit[pnunit.npunch], "#charge exchange rate coefficnt\n" );
		}
	}

	else if( lgMatch("COMP",chCard) )
	{
		/* punch compton, details of the energy exchange problem */
		strcpy( pnunit.chPunch[pnunit.npunch], "COMP" );
		fprintf( pnunit.ipPnunit[pnunit.npunch], "#nu, comup, comdn\n" );
	}

	else if( lgMatch("COOL",chCard) )
	{
		/* cooling, actually done by routine */
		strcpy( pnunit.chPunch[pnunit.npunch], "COOL" );
		fprintf( pnunit.ipPnunit[pnunit.npunch], "#temp\theat\tcool\tfrac coolant \n" );
	}

	/* the punch continuum command, with many options,
	 * the first 3 char of the chPunch flag will always be "CON" 
	 * with the last indicating which one */
	else if( lgMatch("CONT",chCard) )
	{
		/* this flag is checked in PrtComment to generate a caution
		 * if continuum is punched but iterations not performed */
		pnunit.lgPunContinuum = TRUE;

		/* check for keyword UNITS on line, then scan wavelength or energy units if present,
		 * units are copied into pnunit.chConPunEnr */
		ChkUnits(chCard);

		if( lgMatch("BINS",chCard) )
		{
			/* continuum binning */
			strcpy( pnunit.chPunch[pnunit.npunch], "CONB" );

			fprintf( pnunit.ipPnunit[pnunit.npunch], 
				"#Continuum binning, enrOrg, Enr, width of cells\n" );
		}

		else if( lgMatch("DIFF",chCard) )
		{
			/* diffuse continuum */
			strcpy( pnunit.chPunch[pnunit.npunch], "COND" );

			fprintf( pnunit.ipPnunit[pnunit.npunch], 
				"#DiffConEner interactive, non-interactive\n" );
		}

		else if( lgMatch("EMIT",chCard) )
		{
			/* emitted continuum */
			strcpy( pnunit.chPunch[pnunit.npunch], "CONE" );

			fprintf( pnunit.ipPnunit[pnunit.npunch], 
				"#Enr      reflec     outward     total line cont\n" );
		}

		else if( lgMatch("FEII",chCard) )
		{
			/* FeII continuum, only valid if large atom is set */
			strcpy( pnunit.chPunch[pnunit.npunch], "CONF" );

			fprintf( pnunit.ipPnunit[pnunit.npunch], 
				"#FeII: wavelength(A)\tsummed intensity\n" );
		}

		else if( lgMatch("GRAI",chCard) )
		{
			/* grain continuum */
			strcpy( pnunit.chPunch[pnunit.npunch], "CONG" );

			fprintf( pnunit.ipPnunit[pnunit.npunch], 
				"#energy,     graphite,  rest \n" );
		}

		else if( lgMatch("INCI",chCard) )
		{
			/* incident continuum */
			strcpy( pnunit.chPunch[pnunit.npunch], "CONC" );

			fprintf( pnunit.ipPnunit[pnunit.npunch], 
				"#Incident Continuum, Enr nFn \n" );
		}

		else if( lgMatch("INTE",chCard) )
		{
			/* continuum interactions */
			strcpy( pnunit.chPunch[pnunit.npunch], "CONi" );

			fprintf( pnunit.ipPnunit[pnunit.npunch], 
				"#Continuum interactions, inc, otslin. otscon, ConOutInter, outlin \n" );
			/* this is option for lowest energy, if nothing then zero */
			ipFFmt = 3;
			pnunit.punarg[0][pnunit.npunch] = (float)FFmtRead(chCard,&ipFFmt,
			  LINELENGTH,&lgEOL);
		}

		else if( lgMatch("IONI",chCard) )
		{
			/* punch ionizing continuum*/
			strcpy( pnunit.chPunch[pnunit.npunch], "CONI" );

			/* this is option for lowest energy, if nothing then zero */
			ipFFmt = 3;
			pnunit.punarg[0][pnunit.npunch] = (float)FFmtRead(chCard,&ipFFmt,
			  LINELENGTH,&lgEOL);

			/* this is option for smallest interaction to punch, def is 1 percent */
			pnunit.punarg[1][pnunit.npunch] = (float)FFmtRead(chCard,&ipFFmt,LINELENGTH,&lgEOL);
			if( lgEOL )
			{
				pnunit.punarg[1][pnunit.npunch] = 0.01f;
			}
			/* put the header at the top of the file */
			fprintf( pnunit.ipPnunit[pnunit.npunch], 
				"#nu\tflux\tflx*cs\tFinc\totsl\totsc\tout\trate/tot\tintegral\tline\tcont\n" );
		}

		else if( lgMatch("OUTW",chCard) )
		{
			/* outward only continuum */
			if( lgMatch("LOCA",chCard) )
			{
				strcpy( pnunit.chPunch[pnunit.npunch], "CONo" );
				fprintf( pnunit.ipPnunit[pnunit.npunch], 
					"#Local Out   SavOutCon  ConOutInter+line SvOt*opc pass*opc\n" );
			}
			else
			{
				strcpy( pnunit.chPunch[pnunit.npunch], "CONO" );
				fprintf( pnunit.ipPnunit[pnunit.npunch], 
					"#Out Con      OutIncid  OutConD   OutLinD   OutConS\n" );
			}
		}

		else if( lgMatch("TRAN",chCard) )
		{
			/* transmitted continuum */
			strcpy( pnunit.chPunch[pnunit.npunch], "CONT" );

			fprintf( pnunit.ipPnunit[pnunit.npunch], 
				"#Transmitted Continuum\n" );
		}

		else if( lgMatch(" RAW",chCard) )
		{
			/* "raw" continua */
			strcpy( pnunit.chPunch[pnunit.npunch], "CORA" );

			fprintf( pnunit.ipPnunit[pnunit.npunch], 
				"#Raw Con anu\tflux\totslin\totscon\tConRefIncid\tConRefDiff\tConOutIter\tConOutRecInter\toutlin\tConOtNoInter\tThrowOut\tline\tcont\n" );
		}

		else if( lgMatch("REFL",chCard) )
		{
			/* reflected continuum */
			strcpy( pnunit.chPunch[pnunit.npunch], "CONR" );

			fprintf( pnunit.ipPnunit[pnunit.npunch], 
				"#Reflected      cont       line       total        albedo    ConID\n" );
		}

		else
		{
			/* this is the usual punch continuum command */
			strcpy( pnunit.chPunch[pnunit.npunch], "CON " );
			fprintf( pnunit.ipPnunit[pnunit.npunch], 
				"#Cont nu\tincident\ttrans\tDiffOut\tnet trans\treflc\ttotal\tline\tcont\n" );
		}

	}
	else if( lgMatch("CONV",chCard) )
	{
		/* punch reason model declared not converged
		 * not a true punch command, since done elsewhere */
		PunConv.ipPunConv = pnunit.ipPnunit[pnunit.npunch];
		PunConv.lgPunConv = TRUE;
		fprintf( PunConv.ipPunConv, " reason for continued iterations\n" );
		/* this does not count as a punch option (really) */
		pnunit.ipPnunit[pnunit.npunch] = NULL;
		pnunit.npunch -= 1;
	}

	else if( lgMatch(" DR ",chCard) )
	{
		ipdr.ipDRout = pnunit.ipPnunit[pnunit.npunch];
		ipdr.lgDROn = TRUE;

		/* set punch last flag to whatever it was above */
		ipdr.lgDRPLst = pnunit.lgPunLstIter[pnunit.npunch];

		fprintf( ipdr.ipDRout, "zone\tdepth\tdr\tdr 2 go\treason \n" );

		/* this does not count as a punch option (really) */
		pnunit.ipPnunit[pnunit.npunch] = NULL;
		pnunit.npunch -= 1;
	}

	else if( lgMatch("ELEM",chCard) )
	{
		/* option to punch ionization structure of some element */
		strcpy( pnunit.chPunch[pnunit.npunch], "ELEM" );
		GetElem(chCard,&j);
		++j;
		/* this is the atomic number on the physical scale */
		pnunit.punarg[0][pnunit.npunch] = (float)j;
		assert( j>0);
		fprintf( pnunit.ipPnunit[pnunit.npunch], "#depth");
		for(i=0; i<j+1;++i )
		{
			assert( j>0 );
			fprintf( pnunit.ipPnunit[pnunit.npunch], 
				"\t%2s%2li", elementnames.chElementSym[j-1],i+1);
		}
		fprintf( pnunit.ipPnunit[pnunit.npunch], "\n");
	}

	else if( lgMatch("FEII",chCard) )
	{
		if( lgMatch("FRED",chCard) )
		{
			/* punch out some stuff for Fred's FeII atom */
			fprintf( pnunit.ipPnunit[pnunit.npunch], "#FeII rates \n" );
			strcpy( pnunit.chPunch[pnunit.npunch], "FE2f" );
		}
		else if( lgMatch("DEPA",chCard) )
		{
			/* punch out departure coef for the large FeII atom */
			fprintf( pnunit.ipPnunit[pnunit.npunch], "#FeII departure coef \n" );
			/* optional keyword all means do all levels, if not then just do subset */
			if( lgMatch(" ALL",chCard) )
			{
				/* punch all levels, calls routine FeIIPunDepart */
				strcpy( pnunit.chPunch[pnunit.npunch], "FE2D" );
			}
			else
			{
				/* punch a very few selected levels, calls routine FeIIPunDepart */
				strcpy( pnunit.chPunch[pnunit.npunch], "FE2d" );
			}
		}
		else if( lgMatch("POPU",chCard) )
		{
			/* punch out level populations for the large FeII atom */
			fprintf( pnunit.ipPnunit[pnunit.npunch], "#FeII level populations \n" );
			/* optional keyword all means do all levels, if not then just do subset */
			if( lgMatch(" ALL",chCard) )
			{
				/* punch all levels, calls routine FeIIPunDepart */
				strcpy( pnunit.chPunch[pnunit.npunch], "FE2P" );
			}
			else
			{
				/* punch a very few selected levels, calls routine FeIIPunDepart */
				strcpy( pnunit.chPunch[pnunit.npunch], "FE2p" );
			}
		}
		else
		{
			fprintf( ioQQQ, " There must be a second keyword on this command.\n" );
			fprintf( ioQQQ, " The ones I know about are departure and fred.\n" );
			fprintf( ioQQQ, " Sorry.\n" );
			puts( "[Stop in ParsePunch]" );
			cdEXIT(1);
		}
	}

	else if( lgMatch("GAMM",chCard) )
	{
		/* punch all photoionizaiton rates for all subshells */
		fprintf( pnunit.ipPnunit[pnunit.npunch], "#Photoionization rates \n" );
		strcpy( pnunit.chPunch[pnunit.npunch], "GAMM" );

	}
	else if( lgMatch("GRAI",chCard) )
	{
		/* punch grains
		 * chDst = 'NONE' */
		if( lgMatch("OPAC",chCard) )
		{
			strcpy( pnunit.chPunch[pnunit.npunch], "DUSO" );
			/* punch grain opacity command in twice, here and above in opacity
			 * chDst = 'OPAC' */
			fprintf( pnunit.ipPnunit[pnunit.npunch], "#grain ext, abs, scat\n" );
		}
		else if( lgMatch("PHYS",chCard) )
		{
			strcpy( pnunit.chPunch[pnunit.npunch], "DUSP" );
			/* chDst = 'PHYS' */
			fprintf( pnunit.ipPnunit[pnunit.npunch], "#grain temp, charge, drift, frc heat, cool\n" );
		}
		else if( lgMatch(" QS ",chCard) )
		{
			strcpy( pnunit.chPunch[pnunit.npunch], "DUSQ" );
			fprintf( pnunit.ipPnunit[pnunit.npunch], "#grain abs, scat, qs\n" );
		}
		else
		{
			fprintf( ioQQQ, " There must be a second key; they are  OPAC, _QS_, and PHYS\n" );
			puts( "[Stop in ParsePunch]" );
			cdEXIT(1);
		}
	}

	else if( lgMatch("GAUN",chCard) )
	{
		strcpy( pnunit.chPunch[pnunit.npunch], "GAUN" );
		fprintf( pnunit.ipPnunit[pnunit.npunch], "#Gaunt factors.\n" );
	}

	else if( lgMatch("HTWO",chCard) )
	{
		iph2.ipPunH2 = pnunit.ipPnunit[pnunit.npunch];
		iph2.lgPunH2 = TRUE;
		fprintf( pnunit.ipPnunit[pnunit.npunch], "#H2 creation, destruction. \n" );
		/* this does not count as a punch option (really) */
		pnunit.ipPnunit[pnunit.npunch] = NULL;
		pnunit.npunch -= 1;

	}

	/* QHEAT has to be in front of HEAT... */
	else if( lgMatch("QHEA",chCard) )
	{
		gv.QHPunchFile = pnunit.ipPnunit[pnunit.npunch];
		gv.lgQHPunLast = pnunit.lgPunLstIter[pnunit.npunch];
		fprintf( gv.QHPunchFile, "#Probability distributions from quantum heating routine.\n" );
		/* this does not count as a punch option (really) */
		pnunit.ipPnunit[pnunit.npunch] = NULL;
		pnunit.npunch -= 1;
	}

	else if( lgMatch("HEAT",chCard) )
	{
		strcpy( pnunit.chPunch[pnunit.npunch], "HEAT" );
		fprintf( pnunit.ipPnunit[pnunit.npunch], "#\thtot\tctot\theat fracs\n" );
	}

	/* information about the helium atom */
	else if( lgMatch("HELI",chCard) )
	{
		/* singlet populations and rates */
		if( lgMatch("SING",chCard) )
		{
			strcpy( pnunit.chPunch[pnunit.npunch], "HEL1" );
			fprintf( pnunit.ipPnunit[pnunit.npunch], "#he1    level,   simple,  clhe2ov1  smhe2ov1   rec,     ioniz,    hei3 \n" );
		}

		/* triplet populations and rates */
		else if( lgMatch("TRIP",chCard) )
		{
			strcpy( pnunit.chPunch[pnunit.npunch], "HEL3" );
			fprintf( pnunit.ipPnunit[pnunit.npunch], "#HeTrip  level,    simple,    rec,     ioniz,    coll \n" );
		}
		else
		{
			fprintf( ioQQQ, " did not find a keyword on this  line.  Keys are SING, _ION, and TRIP.\n" );
			puts( "[Stop in ParsePunch]" );
			cdEXIT(1);
		}
	}

	else if( lgMatch("HUMM",chCard) )
	{
		strcpy( pnunit.chPunch[pnunit.npunch], "HUMM" );
		fprintf( pnunit.ipPnunit[pnunit.npunch], "#input to DHs routine.\n" );
	}

	else if( lgMatch("HYDR",chCard) )
	{
		/* punch hydrogen physical conditions */
		if( lgMatch("COND",chCard) )
		{
			strcpy( pnunit.chPunch[pnunit.npunch], "HYDc" );
			fprintf( pnunit.ipPnunit[pnunit.npunch], "#depth\tTe\tHDEN\tEDEN\tHI/H\tHII/H\tH2/H\tH2+/H\tH3+/H\tH-/H\n" );
			/* punch hydrogen ionization */
		}
		else if( lgMatch("IONI",chCard) )
		{
			strcpy( pnunit.chPunch[pnunit.npunch], "HYDi" );
			fprintf( pnunit.ipPnunit[pnunit.npunch], "#  gam1    RecEff   HRecRad    hii/hi Shii/hi   hrec(13) dec grd  GamEff   ColEff\n" );
			/* punch hydrogen populations */
		}
		else if( lgMatch("POPU",chCard) )
		{
			strcpy( pnunit.chPunch[pnunit.npunch], "HYDp" );
			fprintf( pnunit.ipPnunit[pnunit.npunch], "#  depth      HI      HII       HN                                                   2s       2p\n" );
		}
		else if( lgMatch("RECC",chCard) )
		{
			/* recombination cooling, for AGN */
			strcpy( pnunit.chPunch[pnunit.npunch], "HYDr" );
			fprintf( pnunit.ipPnunit[pnunit.npunch], "#T\tbAS\tb1\tbB\n" );
		}
		else
		{
			fprintf( ioQQQ, " Punch hydrogen has options: CONDitions POPUlations and IONIzation.\n" );
			puts( "[Stop in ParsePunch]" );
			cdEXIT(1);
		}
	}

	else if( lgMatch("IONI",chCard) )
	{
		/* punch table giving ionization means */
		strcpy( pnunit.chPunch[pnunit.npunch], "IONI" );
		fprintf( pnunit.ipPnunit[pnunit.npunch], 
			"#Mean ionization distribution\n" );
	}

	else if( lgMatch(" IP ",chCard) )
	{
		strcpy( pnunit.chPunch[pnunit.npunch], " IP " );
		fprintf( pnunit.ipPnunit[pnunit.npunch], 
			"#ionization potentials, valence shell\n" );
	}

	else if( lgMatch("LINE",chCard) )
	{
		
		/* scan off energy units for this output */
		ChkUnits(chCard);

		/* options are punch line structure, line intensity, line array,
		 * and line data */
		if( lgMatch("STRU",chCard) )
		{
			/* this will be line structure, emissivity vs depth */
			strcpy( pnunit.chPunch[pnunit.npunch], "LINS" );
			fprintf( pnunit.ipPnunit[pnunit.npunch], 
				"#Emission line structure.\n" );
			/* read in the list of lines to examine */
			punlin(pnunit.ipPnunit[pnunit.npunch],"READ",FALSE);
		}

		else if( lgMatch("CUMU",chCard) )
		{
			/* this will be integrated line intensity, fcn of depth */
			strcpy( pnunit.chPunch[pnunit.npunch], "LINC" );
			/* option for either relative intensity or abs luminosity */
			if( lgMatch("RELA",chCard) )
			{
				lgEOL = TRUE;
				fprintf( pnunit.ipPnunit[pnunit.npunch], 
					"#Cumulative emission line relative intensity.\n" );
			}
			else
			{
				fprintf( pnunit.ipPnunit[pnunit.npunch], 
					"#Cumulative emission line absolute intensity.\n" );
				lgEOL = FALSE;
			}
			/* read in the list of lines to examine */
			punlin(pnunit.ipPnunit[pnunit.npunch],"READ",lgEOL);
		}

		else if( lgMatch("DATA",chCard) )
		{
			/* punch line data, done in DoPunchLineData */
			strcpy( pnunit.chPunch[pnunit.npunch], "LIND" );
			fprintf( pnunit.ipPnunit[pnunit.npunch], "#Emission line data.\n" );
		}

		else if( lgMatch("ARRA",chCard) )
		{
			/* punch energies and luminosities of transferred lines */
			strcpy( pnunit.chPunch[pnunit.npunch], "LINA" );
			fprintf( pnunit.ipPnunit[pnunit.npunch], "#enr\tintensity\tid\ttype\n" );

		}

		else if( lgMatch("LABE",chCard) )
		{
			/* punch line labels */
			strcpy( pnunit.chPunch[pnunit.npunch], "LINL" );
			fprintf( pnunit.ipPnunit[pnunit.npunch], "#Emission line labels\n" );
		}

		else if( lgMatch("OPTI",chCard) )
		{
			/* punch line optical depths */
			strcpy( pnunit.chPunch[pnunit.npunch], "LINO" );
			fprintf( pnunit.ipPnunit[pnunit.npunch], "#species\tenergy(Ryd)\tlog tau\n" );
			/* this is optional limit to smallest optical depths */
			ipFFmt = 5;
			pnunit.punarg[0][pnunit.npunch] = (float)pow(10.,FFmtRead(chCard,&ipFFmt, LINELENGTH,&lgEOL));
			/* this is default of 0.1 napier */
			if( lgEOL )
			{
				pnunit.punarg[0][pnunit.npunch] = 0.1f;
			}
		}

		else if( lgMatch("POPU",chCard) )
		{
			/* punch level populations */
			strcpy( pnunit.chPunch[pnunit.npunch], "LINP" );
			fprintf( pnunit.ipPnunit[pnunit.npunch], "#population information\n" );
			/* this is optional limit to smallest population to punch */
			ipFFmt = 5;
			pnunit.punarg[0][pnunit.npunch] = (float)pow(10.,FFmtRead(chCard,&ipFFmt, LINELENGTH,&lgEOL));
			/* this is default of all populations */
			if( lgEOL )
			{
				pnunit.punarg[0][pnunit.npunch] = 0.f;
			}
		}

		else if( lgMatch("INTE",chCard) )
		{
			/* this will be full set of line intensities */
			strcpy( pnunit.chPunch[pnunit.npunch], "LINI" );
			fprintf( pnunit.ipPnunit[pnunit.npunch], 
				"#Emission line intensities per unit inner area\n" );
			if( lgMatch("COLU",chCard) )
			{
				/* column is key to punch single column */
				strcpy( PunRltType.chPunRltType, "column" );
			}
			else
			{
				/* array is key to punch large array */
				strcpy( PunRltType.chPunRltType, "array " );
			}
			if( lgMatch("EVER",chCard) )
			{
				ipFFmt = 3;
				levery.LinEvery = (long int)FFmtRead(chCard,&ipFFmt,LINELENGTH,&lgEOL);
				levery.lgLinEvery = TRUE;
				if( lgEOL )
				{
					fprintf( ioQQQ, 
						" There must be a second number, the number of zones to print.\n" );
					puts( "[Stop in ParsePunch]" );
					cdEXIT(1);
				}
			}
			else
			{
				levery.LinEvery = ZoneCnt.nend[0];
				levery.lgLinEvery = FALSE;
			}
		}
		else
		{
			fprintf( ioQQQ, 
				" This option for PUNCH LINE is something that I dont understand.  Sorry.\n" );
			puts( "[Stop in ParsePunch]" );
			cdEXIT(1);
		}
	}

	else if( lgMatch("LYMA",chCard) )
	{
		/* lyman alpha lya details */
		strcpy( pnunit.chPunch[pnunit.npunch], "LYMA" );
		fprintf( pnunit.ipPnunit[pnunit.npunch], 
			"#depth\tTau\tN2/N1\tTexc\tTe\tTex/Te\tdest\tots\n" );
	}

	else if( lgMatch(" MAP",chCard) )
	{
		strcpy( pnunit.chPunch[pnunit.npunch], "MAP " );
		fprintf( pnunit.ipPnunit[pnunit.npunch], "#te, heating, cooling.\n" );
		/* do cooling space map for specified zones
		 * if no number, or <0, do map and punch out without doing first zone */
		i = 5;
		MapPar.MapZone = (long)FFmtRead(chCard,&i,LINELENGTH,&lgEOL);
		if( lgEOL )
		{
			MapPar.MapZone = 1;
		}

		/* say output goes to special punch */
		ioMAP = pnunit.ipPnunit[pnunit.npunch];

		if( lgMatch("RANG",chCard) )
		{
			int lgLogOn;
			MapPar.RangeMap[0] = (float)FFmtRead(chCard,&i,LINELENGTH,&lgEOL);
			if( MapPar.RangeMap[0] <= 10. )
			{
				MapPar.RangeMap[0] = (float)pow(10.,MapPar.RangeMap[0]);
				lgLogOn = TRUE;
			}
			else
			{
				lgLogOn = FALSE;
			}

			MapPar.RangeMap[1] = (float)FFmtRead(chCard,&i,LINELENGTH,&lgEOL);
			if( lgLogOn )
				MapPar.RangeMap[1] = (float)pow(10.,MapPar.RangeMap[1]);

			if( lgEOL )
			{
				fprintf( ioQQQ, " There must be a zone number, followed by two temperatures, on this line.  Sorry.\n" );
				puts( "[Stop in ParsePunch]" );
				cdEXIT(1);
			}
		
		}
	}

	else if( lgMatch("MOLE",chCard) )
	{
		/* molecules, especially for pdr calculations */
		strcpy( pnunit.chPunch[pnunit.npunch], "MOLE" );
		fprintf( pnunit.ipPnunit[pnunit.npunch], "#dr\tH2/H\tH2*/H\tCO/H\tH2O/H\tOH/H\tCH/HO2/H\n" );
	}

	else if( lgMatch("OPTI",chCard) )
	{
		/* check for keyword UNITS on line, then scan wavelength or energy units if present,
		 * units are copied into pnunit.chConPunEnr */
		ChkUnits(chCard);
		/* punch continuum optical depths */
		strcpy( pnunit.chPunch[pnunit.npunch], "OPTI" );
		fprintf( pnunit.ipPnunit[pnunit.npunch], "#energy\ttotal\tabsorp\tscat\n" );

	}
	else if( lgMatch(" OTS",chCard) )
	{
		strcpy( pnunit.chPunch[pnunit.npunch], " OTS" );
		fprintf( pnunit.ipPnunit[pnunit.npunch], "#otscon, lin, conOpac LinOpc\n" );
	}

	else if( lgMatch("OVER",chCard) )
	{
		/* punch overview of model results */
		strcpy( pnunit.chPunch[pnunit.npunch], "OVER" );
		fprintf( pnunit.ipPnunit[pnunit.npunch], 
			"#depth\tTe\tHtot\thden\teden\tHI\tHII\tHeI\tHeII\tHeIII\tC1\tC2\tC3\tC4\tO1\tO2\tO3\tO4\tO5\tO6\n" );
	}

	else if( lgMatch(" PDR",chCard) )
	{
		strcpy( pnunit.chPunch[pnunit.npunch], " PDR" );
		fprintf( pnunit.ipPnunit[pnunit.npunch], 
			"#depth\tH colden\tAV\tTe\tHI/HDEN\tH2/HDEN\tH2*/HDEN\tCI/C\tCO/C\tH2O/O\n" );
	}

	else if( lgMatch("PHYS",chCard) )
	{
		/* punch physical conditions */
		strcpy( pnunit.chPunch[pnunit.npunch], "PHYS" );
		fprintf( pnunit.ipPnunit[pnunit.npunch], 
			"#PhyC    Depth       Te,       HDEN,       Ne,       HTOT,      accel\n" );
	}

	else if( lgMatch("POIN",chCard) )
	{
		/* punch out the pointers */
		PunPoint.ipPoint = pnunit.ipPnunit[pnunit.npunch];
		PunPoint.lgPunPoint = TRUE;
		fprintf( pnunit.ipPnunit[pnunit.npunch], "#pointers. \n" );
		/* this does not count as a punch option (really) */
		pnunit.ipPnunit[pnunit.npunch] = NULL;
		pnunit.npunch -= 1;
	}

	else if( lgMatch("PRES",chCard) )
	{
		strcpy( pnunit.chPunch[pnunit.npunch], "PRES" );
		fprintf( pnunit.ipPnunit[pnunit.npunch], 
			"# P depth,   p(G+L)    PIn+pintg  pgas(r0)   pgas     pram       prad      integ,    windv,     cad,  conv?\n" );
	}

	else if( lgMatch("RADI",chCard) )
	{
		strcpy( pnunit.chPunch[pnunit.npunch], "RADI" );
		fprintf( pnunit.ipPnunit[pnunit.npunch], "#NZONE, radius, r-r0, dr.\n" );
	}

	else if( lgMatch("RECO",chCard) )
	{
		if( lgMatch("COEF",chCard) )
		{
			/* recombination coefficients for everything
			 * ioRecom set to zero in routine zero, non-zero value
			 * is flag to punch recombination coefficients. the output is actually
			 * produced by a series of routines, as they generate the recombination
			 * coefficients.  these include 
			 * dielsupres, helium, hydrorecom, iibod, and makerecomb*/
			ioRecom = pnunit.ipPnunit[pnunit.npunch];
			lgioRecom = TRUE;
			fprintf( pnunit.ipPnunit[pnunit.npunch], 
				"#recombination coefficients cm3 s-1 for current density and temperature\n" );

			/* this does not count as a punch option (really) */
			pnunit.ipPnunit[pnunit.npunch] = NULL;
			pnunit.npunch -= 1;
		}

		else if( lgMatch(" AGN",chCard) )
		{
			/* punch recombination rates for agn table */
			strcpy( pnunit.chPunch[pnunit.npunch], "RECA" );
			fprintf( pnunit.ipPnunit[pnunit.npunch], "#Recomrates for agn table\n" );
		}

		else if( lgMatch("EFFI",chCard) )
		{
			/* punch recombination efficiency */
			strcpy( pnunit.chPunch[pnunit.npunch], "RECE" );
			fprintf( pnunit.ipPnunit[pnunit.npunch], "#Recomn effic H, Heo, He+\n" );
		}

		else
		{
			fprintf( ioQQQ, " No option recognized on this punch recombination command\n" );
			fprintf( ioQQQ, " Valid options are COEFFICIENTS, AGN, and EFFICIENCY\n" );
			puts( "[Stop in ParsePunch]" );
			cdEXIT(1);
		}
	}

	/* punch results command, either as single column or wide array */
	else if( lgMatch("RESU",chCard) )
	{
		strcpy( pnunit.chPunch[pnunit.npunch], "RESU" );
		if( lgMatch("COLU",chCard) )
		{
			/* column is key to punch single column */
			strcpy( PunRltType.chPunRltType, "column" );
		}
		else
		{
			/* array is key to punch large array */
			strcpy( PunRltType.chPunRltType, "array " );
		}

		/* do not change following, is used as flag in getlines */
		fprintf( pnunit.ipPnunit[pnunit.npunch], "#results of calculation\n" );
	}

	else if( lgMatch("SOUR",chCard) )
	{
		if( lgMatch("DEPT",chCard) )
		{
			/* print continuum source function as function of depth */
			strcpy( pnunit.chPunch[pnunit.npunch], "SOUD" );
			fprintf( pnunit.ipPnunit[pnunit.npunch], "#continuum source function vs depth\n" );
		}
		else if( lgMatch("SPEC",chCard) )
		{
			/* print spectrum continuum source function at 1 depth */
			strcpy( pnunit.chPunch[pnunit.npunch], "SOUS" );
			fprintf( pnunit.ipPnunit[pnunit.npunch], "#continuum source function\n" );
		}
		else
		{
			fprintf( ioQQQ, " A second keyword must appear on this line.\n" );
			fprintf( ioQQQ, " They are DEPTH and SPECTRUM.\n" );
			fprintf( ioQQQ, " Sorry.\n" );
			puts( "[Stop in ParsePunch]" );
			cdEXIT(1);
		}
	}

	else if( lgMatch("SPEC",chCard) )
	{
		/* punch special, will call routine PunSpec */
		strcpy( pnunit.chPunch[pnunit.npunch], "SPEC" );
		fprintf( pnunit.ipPnunit[pnunit.npunch], "#Special.\n" );
	}

	else if( lgMatch("TEGR",chCard) )
	{
		/* punch tegrid */
		strcpy( pnunit.chPunch[pnunit.npunch], "TEGR" );
		fprintf( pnunit.ipPnunit[pnunit.npunch], "#zone, te, heat, cool.\n" );
	}

	else if( lgMatch("TEMP",chCard) )
	{
		strcpy( pnunit.chPunch[pnunit.npunch], "TEMP" );
		fprintf( pnunit.ipPnunit[pnunit.npunch], "#DEPTH, te, dt/dr.\n" );
	}

	else if( lgMatch("TPRE",chCard) )
	{
		/* debug output from the temperature predictor in zonestart, 
		 * set with punch tpred command */
		strcpy( pnunit.chPunch[pnunit.npunch], "TPRE" );
		fprintf( pnunit.ipPnunit[pnunit.npunch], 
			"#zone  old temp,  guess Tnew, new temp    delta \n" );
	}

	else if( lgMatch("VERN",chCard) )
	{
		/* punch verner command , three optional parameters, threshold for faintest
		 * line to print, lower and upper energy bounds */

		/* punch intensities from large FeII atom */
		strcpy( pnunit.chPunch[pnunit.npunch], "VERN" );

		/* short keyword makes punch half as big */
		if( lgMatch("SHOR",chCard) )
		{
			FeII.lgShortFe2 = TRUE;
		}
		else
		{
			FeII.lgShortFe2 = FALSE;
		}

		/* first optional number changes the theshold of weakest line to print*/
		i = 5;
		/* fe2thresh is intensity relative to normalization line,
		 * normally Hbeta, and is set to zero in zero.c */
		FeII.fe2thresh = (float)FFmtRead(chCard,&i,LINELENGTH,&lgEOL);
		if( lgEOL )
		{
			FeII.fe2thresh = 0.;
		}

		/* it is a log if negative */
		if( FeII.fe2thresh < 0. )
		{
			FeII.fe2thresh = (float)pow(10.,FeII.fe2thresh);
		}

		/* check for energy range of lines to be punched,
		 * this is to limit size of output file */
		FeII.fe2ener[0] = (float)FFmtRead(chCard,&i,LINELENGTH,&lgEOL);
		if( lgEOL )
		{
			FeII.fe2ener[0] = 0.;
		}

		FeII.fe2ener[1] = (float)FFmtRead(chCard,&i,LINELENGTH,&lgEOL);
		if( lgEOL )
		{
			FeII.fe2ener[1] = 1e8;
		}
		/* if either is negative then both are logs */
		if( FeII.fe2ener[0] < 0. || FeII.fe2ener[1] < 0. )
		{
			FeII.fe2ener[0] = (float)pow(10.,FeII.fe2ener[0]);
			FeII.fe2ener[1] = (float)pow(10.,FeII.fe2ener[1]);
		}
		/* entered in Ryd in above, convert to wavenumbers */
		FeII.fe2ener[0] /= (float)WAVNRYD;
		FeII.fe2ener[1] /= (float)WAVNRYD;

		/* these results are actually created by the FeIIPunchLines routine
		 * that lives in the pop371 file */
		fprintf( pnunit.ipPnunit[pnunit.npunch], "#Results from Verner FeII atom.\n" );
	}

	else if( lgMatch("WIND",chCard) )
	{
		strcpy( pnunit.chPunch[pnunit.npunch], "WIND" );
		fprintf( pnunit.ipPnunit[pnunit.npunch], "#radius, DEPTH,vel, accel, force mul\n" );
	}

	else
	{
		fprintf( ioQQQ, 
			" ParsePunch cannot find a recognized keyword on this PUNCH command line.\n" );
		puts( "[Stop in ParsePunch]" );
		cdEXIT(1);
	}

	/* increment total number of punch commands, */
	++pnunit.npunch;

#	ifdef DEBUG_FUN
	fputs( " <->ParsePunch()\n", debug_fp );
#	endif
	return;
}

/* initialize punch file pointers. KEEP THIS ROUTINE SYNCHED UP WITH THE NEXT ONE */
void InitPunchFiles(void)
{
	long int i;

#	ifdef DEBUG_FUN
	fputs( "<+>InitPunchFiles()\n", debug_fp );
#	endif

	for( i=0; i < LIMPUN; i++ )
	{
		pnunit.ipPnunit[i] = NULL;
	}
	PunConv.ipPunConv = NULL;
	PunConv.lgPunConv = FALSE;
	ipdr.ipDRout = NULL;
	ipdr.lgDROn = FALSE;
	iph2.ipPunH2 = NULL;
	iph2.lgPunH2 = FALSE;
	PunPoint.ipPoint = NULL;
	PunPoint.lgPunPoint = FALSE;
	gv.QHPunchFile = NULL;
	ioRecom = NULL;
	lgioRecom = FALSE;
	ioMAP = NULL;

#	ifdef DEBUG_FUN
	fputs( " <->InitPunchFiles()\n", debug_fp );
#	endif
	return;
}

/* close all punch files. KEEP THIS ROUTINE SYNCHED UP WITH THE PREVIOUS ONE */
void ClosePunchFiles(void)
{
	long int i;

#	ifdef DEBUG_FUN
	fputs( "<+>ClosePunchFiles()\n", debug_fp );
#	endif

	/* close all punch units cloudy opened with punch command */
	for( i=0; i < pnunit.npunch; i++ )
	{
		if( pnunit.ipPnunit[i] != NULL )
		{
			fclose( pnunit.ipPnunit[i] );
			pnunit.ipPnunit[i] = NULL;
		}
	}

	if( PunConv.ipPunConv != NULL )
	{
		fclose( PunConv.ipPunConv );
		PunConv.ipPunConv = NULL;
		PunConv.lgPunConv = FALSE;
	}
	if( ipdr.ipDRout != NULL )
	{
		fclose( ipdr.ipDRout );
		ipdr.ipDRout = NULL;
		ipdr.lgDROn = FALSE;
	}
	if( iph2.ipPunH2 != NULL )
	{
		fclose( iph2.ipPunH2 );
		iph2.ipPunH2 = NULL;
		iph2.lgPunH2 = FALSE;
	}
	if( PunPoint.ipPoint != NULL )
	{
		fclose( PunPoint.ipPoint );
		PunPoint.ipPoint = NULL;
		PunPoint.lgPunPoint = FALSE;
	}
	if( gv.QHPunchFile != NULL )
	{
		fclose( gv.QHPunchFile );
		gv.QHPunchFile = NULL;
	}
	if( ioRecom != NULL )
	{
		fclose( ioRecom );
		ioRecom = NULL;
		lgioRecom = FALSE;
	}
	/* this file was already closed as a pnunit.ipPnunit, erase copy of pointer as well */
	ioMAP = NULL;

#	ifdef DEBUG_FUN
	fputs( " <->ClosePunchFiles()\n", debug_fp );
#	endif
	return;
}

/* check for keyword UNITS on line, then scan wavelength or energy units if present,
 * units are copied into pnunit.chConPunEnr */
static void ChkUnits( char *chCard)
{

#	ifdef DEBUG_FUN
	fputs( "<+>ChkUnits()\n", debug_fp );
#	endif

	/* option to set units for continuum energy in punch output */
	if( lgMatch("NITS",chCard) )
	{
		if( lgMatch("MICR",chCard) )
		{
			strcpy( pnunit.chConPunEnr[pnunit.npunch], "micr" );
		}
		else if( lgMatch(" KEV",chCard) )
		{
			strcpy( pnunit.chConPunEnr[pnunit.npunch], " kev" );
		}
		else if( lgMatch(" EV ",chCard) )
		{
			strcpy( pnunit.chConPunEnr[pnunit.npunch], " ev " );
		}
		else if( lgMatch("ANGS",chCard) )
		{
			strcpy( pnunit.chConPunEnr[pnunit.npunch], "angs" );
		}
		else if( lgMatch("RYDB",chCard) )
		{
			strcpy( pnunit.chConPunEnr[pnunit.npunch], "rydb" );
		}
		else
		{
			fprintf( ioQQQ, " I did not recognize units on this line.\n" );
			fprintf( ioQQQ, " Units are keV, eV, Angstroms, microns\n" );
			puts( "[Stop in ParsePunch]" );
			cdEXIT(1);
		}
	}
	else
	{
		strcpy( pnunit.chConPunEnr[pnunit.npunch], "ryd " );
	}

#	ifdef DEBUG_FUN
	fputs( " <->ChkUnits()\n", debug_fp );
#	endif
	return;
}
