/*PressureTotal determine the gas and line radiation pressures for current conditions,
 * return value is total pressure */
/*RamPressure determine the ram pressure for current conditions */
#include <limits.h>
#include "cddefines.h"
#include "taulines.h"
#include "physconst.h"
#include "hydrogenic.h"
#include "wind.h"
#include "phycon.h"
#include "hmi.h"
#include "hevmolec.h"
#include "ionfracs.h"
#include "trace.h"
#include "atomcwgt.h"
#include "doppvel.h"
#include "ph1com.h"
#include "widla.h"
#include "widthline.h"
#include "nsshells.h"
#include "texcline.h"
#include "ionrange.h"
#include "chlinelbl.h"
#include "pressure.h"

/* return value is total pressure */
double PressureTotal(void)
{
	long int i, 
		ion,
		/* will be used to point to line causing radiation pressure */
	  ip1=LONG_MAX,
	  ip2=LONG_MAX,
	  ip3=LONG_MAX, 
	  ipHi, 
	  ipLo, 
	  ipSet , /* used in debug print statement to see which line adds most pressure */
	  ipZ,
	  nelem;

	double 
		HLike=0.,
		HeLike=0.,
		PLevel1=0.,
		PLevel2=0.,
		PCO=0.,
	  RadPres1, 
	  TotalPressure_v, 
	  p2, 
	  pmx;

#	ifdef DEBUG_FUN
	fputs( "<+>PressureTotal()\n", debug_fp );
#	endif

	/* PressureTotal - evaluate total pressure
	 * this function evaluates the following
	 * PressureTotal - gas and radiation pressure, dynes/cm^2
	 * GasPres gas pressure, same units
	 *
	 * HDEN is defined as total hydrogen nuclei phycon; all forms
	 *
	 * phycon.TotalNuclei is total number of nuclei and molecules
	 * nelem,0 is abundance of nelem */

	/* find total number of nuclei, and total ionization energy of gas */
	phycon.TotalNuclei = 0.;
	phycon.EnergyIonization = 0.;
	for( nelem=0; nelem < LIMELM; nelem++ )
	{
		phycon.TotalNuclei += xIonFracs[0][nelem];
		for( ion=IonRange.IonLow[nelem]; ion<=IonRange.IonHigh[nelem]; ++ion )
		{
			for( i=1; i<=ion; ++i )
			{
				long int nelec = nelem+1 - i ;
				phycon.EnergyIonization += xIonFracs[ion+1][nelem] * 
					PH1COM.PH1[nsShellsCom.nsShells[i-1][nelem]-1][nelec][nelem][0]/EVRYD* 0.9998787;
			}
		}
	}

	/* convert to ergs/cm^3 */
	phycon.EnergyIonization *= EN1RYD;

	/* add on molecules */
	for( i=0; i < NHEVML; i++ )
	{
		phycon.TotalNuclei += hevmolec.hevmol[i];
	}

	/* do not double count C+, which is in this network as ipCP */
	phycon.TotalNuclei -= hevmolec.hevmol[IPCP-1];

	/* add on hydrogen molecules */
	phycon.TotalNuclei += hmi.htwo + hmi.h2plus + hmi.hminus + hmi.h3plus;

	/* phycon.ElecFrac is electron fraction, used for secondary ionization efficiency */
	phycon.ElecFrac = (float)(phycon.eden/phycon.TotalNuclei);

	/* pden is the total number of particles per unit vol */
	phycon.pden = (float)(phycon.eden + phycon.TotalNuclei);

	/* this is total neutral particle phycon for collisional ionization */
	phycon.CollidDensity = 0.;
	for( i=0; i < LIMELM; i++ )
	{
		phycon.CollidDensity += xIonFracs[1][i];
	}

	/* now add all the heavy molecules */
	for( i=0; i < NHEVML; i++ )
	{
		phycon.CollidDensity += hevmolec.hevmol[i];
	}

	/* do not double count C+, which is in this network as ipCP */
	phycon.CollidDensity -= hevmolec.hevmol[IPCP-1];

	phycon.CollidDensity += (float)(hmi.h2plus + hmi.hminus + hmi.h3plus + 
	  2.*hmi.htwo);

	/* gas pressure */
	pressure.GasPres = (float)(phycon.pden*phycon.te*BOLTZMANN);

	/* WMOLE is molecular weight, AtomicMassUnit per particle */
	phycon.wmole = 0.;
	for( i=0; i < LIMELM; i++ )
	{
		phycon.wmole += xIonFracs[0][i]*AtomcWgt.AtomicWeight[i];
	}

	phycon.wmole += (float)(hevmolec.hevmol[IPCH-1]*13. + hevmolec.hevmol[IPCHP-1]*
	  13. + hevmolec.hevmol[IPOH-1]*17. + hevmolec.hevmol[IPOHP-1]*
	  17. + hevmolec.hevmol[IPOTWO-1]*32. + hevmolec.hevmol[IPCTWO-1]*
	  24. + hevmolec.hevmol[IPCO-1]*28. + hevmolec.hevmol[IPCOP-1]*
	  28. + hevmolec.hevmol[IPH2O-1]*18. + hevmolec.hevmol[IPH2OP-1]*
	  18. + hevmolec.hevmol[IPO2P-1]*32. + hevmolec.hevmol[IPC2P-1]*
	  24. + hevmolec.hevmol[IPH3P-1]*3. + hevmolec.hevmol[IPH3OP-1]*
	  19. + hevmolec.hevmol[IPCH2P-1]*14. + hevmolec.hevmol[IPCH2-1]*
	  14. + hevmolec.hevmol[IPCH3-1]*15.);

	phycon.wmole /= phycon.pden;

	/* this is Lya excitation temperature, has always been evaluated here */
	hydro.TexcLya = (float)TexcLine( &HydroLines[ipHYDROGEN][ipH2p][ipH1s] );

	/* xMassDensity is density in grams per cc */
	phycon.xMassDensity = (float)(phycon.wmole*ATOMIC_MASS_UNIT*phycon.pden);

	/* RadPres is pressure due to lines, lgRadPresON turns off or on */
	pressure.RadPres = 0.;
	pmx = 0.;

	/* this is the total Lya line width, in velocity units, cm/s
	 * >>chng 00 dec 04, this had been inside if( pressure.lgRadPrsON but
	 * must always be evaluated since needed for pumping of FeII by Lya */
	/* if statement stops evaluation before initial quantities set */
	if( HydroLines[ipHYDROGEN][ipH2p][ipH1s].PopOpc > 0. )
	{
		hydro.HLineWidth = (float)(
			widla(
			HydroLines[ipHYDROGEN][ipH2p][ipH1s].TauIn,
		   HydroLines[ipHYDROGEN][ipH2p][ipH1s].TauTot,
		   4.72e-2/phycon.sqrte,
			DoppVel.doppler[ipHYDROGEN]));
	}
	else
	{
		hydro.HLineWidth = 0.;
	}

	/* find max rad pressure even if capped
	 * lgRadPrsON is turned off by NO RADIATION PRESSURE command */
	if( pressure.lgRadPrsON )
	{
		ipZ = ipHYDROGEN;
		ipLo = ipH1s;
		ipHi = ipH2p;
		if( HydroLines[ipZ][ipHi][ipLo].PopHi > SMALLFLOAT2X &&
			HydroLines[ipZ][ipHi][ipLo].PopOpc > SMALLFLOAT2X )
		{
			/* this is the 2p/1s level population, will be used for excitation
			 * temperature and radiation pressure */
			p2 = HydroLines[ipZ][ipHi][ipLo].PopHi/HydroLines[ipZ][ipHi][ipLo].PopOpc/3.;

			/* coefficient assumes that HLineWidth is half width, not total */
			pressure.HLalpha = (float)(p2*5.08e-6*hydro.HLineWidth);
		}
		else
		{
			p2 = 0.;
			pressure.HLalpha = 0.;
		}
		pressure.RadPres = pressure.HLalpha;
		/*fprintf(ioQQQ,"presbugg %e\n", pressure.HLalpha);*/

		/* these identify La as the pressure we have so far */
		pmx = pressure.HLalpha;
		ip1 = ipH2p;
		ip2 = ipH1s;
		ip3 = ipHYDROGEN;
		ipSet = 2;
		HLike = 0.;
		/* line radiation pressure for hydrogen lines */
		for( ipZ=0; ipZ < LIMELM; ipZ++ )
		{
			if( IonRange.IonHigh[ipZ] == ipZ + 1 )
			{
				/* >>chng 01 jan 08, do not include highest level since maser can occur,
				 * and pops are set to small number in this case */
				for( ipLo=ipH1s; ipLo < (nHLevels[ipZ] - 1); ipLo++ ) 
				{
					/* this loop does not include Lya since done just above, 
					 * and set into plints[0] */
					for( ipHi=MAX2( 3 , ipLo+1 ); ipHi < nHLevels[ipZ]; ipHi++ )
					{
						if( HydroLines[ipZ][ipHi][ipLo].PopHi > SMALLFLOAT2X &&
							HydroLines[ipZ][ipHi][ipLo].PopOpc > SMALLFLOAT2X )
						{
							RadPres1 = 5.551e-2*
							  (powi(HydroLines[ipZ][ipHi][ipLo].EnergyWN/1.e6,4))*
							  (HydroLines[ipZ][ipHi][ipLo].PopHi/
							  HydroLines[ipZ][ipHi][ipLo].gHi)/
							  (HydroLines[ipZ][ipHi][ipLo].PopOpc/
							  HydroLines[ipZ][ipHi][ipLo].gLo)*
							  /* WidthLine gets linewidth in terms of RT effects */
							  WidthLine(&HydroLines[ipZ][ipHi][ipLo])* 2.;
							if( RadPres1 > pmx )
							{
								ipSet = 2;
								pmx = RadPres1;
								ip1 = ipHi;
								ip2 = ipLo;
								ip3 = ipZ;
							}
							HLike += RadPres1;
							pressure.RadPres += (float)RadPres1;
						}
					}
				}
			}
		}

		HeLike = 0.;
		/* line radiation pressure for he-like lines */
		for( ipZ=1; ipZ < LIMELM; ipZ++ )
		{
			if( IonRange.IonHigh[ipZ] >= ipZ )
			{
				for( ipLo=0; ipLo < (nHeLevels[ipZ] - 1); ipLo++ ) 
				{
					for( ipHi=ipLo+1; ipHi <nHeLevels[ipZ]; ipHi++ )
					{
						if( HeLines[ipZ][ipHi][ipLo].PopHi > SMALLFLOAT2X &&
							HeLines[ipZ][ipHi][ipLo].PopOpc > SMALLFLOAT2X )
						{
							RadPres1 = 5.551e-2*
							  (powi(HeLines[ipZ][ipHi][ipLo].EnergyWN/1.e6,4))*
							  (HeLines[ipZ][ipHi][ipLo].PopHi/
							  HeLines[ipZ][ipHi][ipLo].gHi)/
							  (HeLines[ipZ][ipHi][ipLo].PopOpc/
							  HeLines[ipZ][ipHi][ipLo].gLo)*
							  /* WidthLine gets linewidth in terms of RT effects */
							  WidthLine(&HeLines[ipZ][ipHi][ipLo])* 2.;
							if( RadPres1 > pmx )
							{
								ipSet = 7;
								pmx = RadPres1;
								ip1 = ipHi;
								ip2 = ipLo;
								ip3 = ipZ;
							}
							HeLike += (float)RadPres1;
							pressure.RadPres += (float)RadPres1;
						}
					}
				}
			}
		}

		/* >>chng 01 jan 08, kill this block */
#		if 0
		/* treat hydrogen's lya special, since it is special,
		 * but only do this if hydrogen is ionized */
		if( IonRange.IonHigh[0] == 1 )
		{
			/* hydrogen Lya by itself */
			if( HydroLines[ipHYDROGEN][ipH2p][ipH1s].PopHi > 1e-30 )
			{
				RadPres1 = 5.551e-2*(powi(HydroLines[ipHYDROGEN][ipH2p][ipH1s].EnergyWN/
				  1.e6,4))*(HydroLines[ipHYDROGEN][ipH2p][ipH1s].PopHi/
				  HydroLines[ipHYDROGEN][ipH2p][ipH1s].gHi)/
				  (HydroLines[ipHYDROGEN][ipH2p][ipH1s].PopLo/
				  HydroLines[ipHYDROGEN][ipH2p][ipH1s].gLo)*
				  WidthLine(&HydroLines[ipHYDROGEN][ipH2p][ipH1s])* 2.;
			}
			else
			{
				RadPres1 = 0.;
			}

			/* correct for counting Lya radiation pressure twice */
			pressure.RadPres = (float)MAX2(0.,pressure.RadPres-RadPres1);
		}
#		endif

		PLevel1 = 0.;
		/* line radiation pressure from large set of level 1 lines */
		for( i=1; i <= nLevel1; i++ )
		{
			if( TauLines[i].PopHi > 1e-30 )
			{
				RadPres1 = 5.551e-2*(powi(TauLines[i].EnergyWN/
				  1.e6,4))*(TauLines[i].PopHi/TauLines[i].gHi)/
				  (TauLines[i].PopLo/TauLines[i].gLo)*
				  WidthLine(&TauLines[i])*2.;
				if( RadPres1 > pmx )
				{
					ipSet = 3;
					pmx = RadPres1;
					ip1 = i;
				}
				PLevel1 += (float)RadPres1;
				pressure.RadPres += (float)RadPres1;
			}
		}

		/* level 2 lines */
		PLevel2 = 0.;
		for( i=0; i < nWindLine; i++ )
		{
			if( TauLine2[i].PopHi > 1e-30 )
			{
				RadPres1 = 5.551e-2*(powi(TauLine2[i].EnergyWN/
				  1.e6,4))*(TauLine2[i].PopHi/TauLine2[i].gHi)/
				  (TauLine2[i].PopLo/TauLine2[i].gLo)*
				  WidthLine(&TauLine2[i])*2.;
				pressure.RadPres += (float)RadPres1;
				PLevel2 += (float)RadPres1;
				if( RadPres1 > pmx )
				{
					ipSet = 4;
					pmx = RadPres1;
					ip1 = i;
				}
			}
		}

		PCO = 0.;
		/* co carbon monoxide lines */
		for( i=0; i < nCORotate; i++ )
		{
			if( C12O16Rotate[i].PopHi > 1e-30 )
			{
				RadPres1 = 5.551e-2*(powi(C12O16Rotate[i].EnergyWN/
				  1.e6,4))*(C12O16Rotate[i].PopHi/C12O16Rotate[i].gHi)/
				  (C12O16Rotate[i].PopLo/C12O16Rotate[i].gLo)*
				  WidthLine(&C12O16Rotate[i])*2.;
				pressure.RadPres += (float)RadPres1;
				PCO += (float)RadPres1;
				if( RadPres1 > pmx )
				{
					ipSet = 5;
					pmx = RadPres1;
					ip1 = i;
				}
			}
			if( C13O16Rotate[i].PopHi > 1e-30 )
			{
				RadPres1 = 5.551e-2*(powi(C13O16Rotate[i].EnergyWN/
				  1.e6,4))*(C13O16Rotate[i].PopHi/C13O16Rotate[i].gHi)/
				  (C13O16Rotate[i].PopLo/C13O16Rotate[i].gLo)*
				  WidthLine(&C13O16Rotate[i])*2.;
				pressure.RadPres += (float)RadPres1;
				PCO += (float)RadPres1;
				if( RadPres1 > pmx )
				{
					ipSet = 6;
					pmx = RadPres1;
					ip1 = i;
				}
			}
		}

	}
	else
	{
		/* case where radiation pressure is not turned on */
		ip1 = 0;
		ipSet = 0;
	}

	/* PBETA is ratio of radiation to gas + continuum pressure */
	if( pressure.lgConPres )
	{
		/* this includes integrated incident continuum */
		pressure.pbeta = pressure.RadPres/(pressure.GasPres + pressure.PresInteg);
	}
	else
	{
		/* this does not */
		pressure.pbeta = pressure.RadPres/pressure.GasPres;
	}

	{
		/* option to print particulars of some line when called */
		/*@-redef@*/
		enum {DEBUG=FALSE};
		/*@+redef@*/
		if( DEBUG )
		{
			if( pressure.pbeta >0.2 && iteration > 2 )
			{
				fprintf(ioQQQ,
					"PressureTotal zone %li Prad/Pgas, constituents follow H%e He%e L1%e L2%e CO%e \n",
					nzone,
				HLike,
				HeLike,
				PLevel1,
				PLevel2,
				PCO);
			}
		}
	}

	/* this would be a major logical error .. .. . */
	if( pressure.RadPres < 0. )
	{
		fprintf(ioQQQ,"negative pressure, constituents follow %e %e %e %e %e \n",
		HLike,
		HeLike,
		PLevel1,
		PLevel2,
		PCO);
		ShowMe();
		puts( "[Stop in PressureTotal]" );
		cdEXIT(1);
	}

	/* following test will never suceed, here to trick lint, ipSet is only used
	 * when needed for debug */
	if( trace.lgTrace && ipSet <0 )
	{
		fprintf(ioQQQ,
			"      PressureTotal, pressure.pbeta = %e, ipSet%li ip1=%li \n", 
			pressure.pbeta,ipSet,ip1 );
	}

	/* stop model from running away on first iteration, when line optical
	 * depths are not defined correctly anyway.
	 * if( iter.le.itermx .and. RadPres.ge.GasPres ) then
	 * >>chng 97 jul 23, only cap radiation pressure on first iteration */
	if( iteration <= 1 && pressure.RadPres >= pressure.GasPres )
	{
		pressure.RadBetaMax = (float)MAX2(pressure.RadBetaMax,pressure.pbeta);
		/* stop RadPres from exceeding the gas pressure on first iteration */
		pressure.RadPres = (float)MIN2(pressure.RadPres,pressure.GasPres);
		pressure.lgPradCap = TRUE;
	}

	/* radiation pressure only included in total eqn of state when this flag is
	 * true, set with constant pressure command */
	if( pressure.lgRadPresON )
	{
		TotalPressure_v = pressure.GasPres + pressure.RadPres;
	}
	else
	{
		TotalPressure_v = pressure.GasPres;
	}

	/* remember the globally most important line, in the entire model */
	if( pressure.pbeta > pressure.RadBetaMax )
	{
		pressure.RadBetaMax = pressure.pbeta;
		pressure.ipPradMax = ip1;
		if( ipSet == 1 )
		{
			/* special first hydrogenic set */
			strcpy( pressure.chLineRadPres, "First 20" );
		}
		else if( ipSet == 2 )
		{
			/* hydrogenic */
			strcpy( pressure.chLineRadPres, "Hydro ");
			strcat( pressure.chLineRadPres, chLineLbl(&HydroLines[ip3][ip1][ip2]) );
		}
		else if( ipSet == 3 )
		{
			/* level 1 lines */
			strcpy( pressure.chLineRadPres, "Level1 ");
			strcat( pressure.chLineRadPres, chLineLbl(&TauLines[ip1]) );
		}
		else if( ipSet == 4 )
		{
			/* level 2 lines */
			strcpy( pressure.chLineRadPres, "Level2 ");
			strcat( pressure.chLineRadPres, chLineLbl(&TauLine2[ip1]) );
		}
		else if( ipSet == 5 )
		{
			/* c12o16 carbon monoxide lines */
			strcpy( pressure.chLineRadPres, "12CO ");
			strcat( pressure.chLineRadPres, chLineLbl(&C12O16Rotate[ip1]) );
		}
		else if( ipSet == 6 )
		{
			/* c13o16 carbon monoxide lines */
			strcpy( pressure.chLineRadPres, "13CO ");
			strcat( pressure.chLineRadPres, chLineLbl(&C13O16Rotate[ip1]) );
		}
		else if( ipSet == 7 )
		{
			/* helium lines */
			strcpy( pressure.chLineRadPres, "HeLike ");
			strcat( pressure.chLineRadPres, chLineLbl(&HeLines[ip3][ip1][ip2]) );
		}
		else
		{
			fprintf(ioQQQ," PressureTotal ipSet insanity.  this is impossible.\n");
			ShowMe();
			puts( "[Stop in PressureTotal]" );
			cdEXIT(1);
		}
	}

	/* remember highest pressure anywhere */
	pressure.PresMax = MAX2(pressure.PresMax,(float)TotalPressure_v);


#	ifdef DEBUG_FUN
	fputs( " <->PressureTotal()\n", debug_fp );
#	endif
	return( TotalPressure_v );
}

double RamPressure(void)
{
	double RamPressure_v;

#	ifdef DEBUG_FUN
	fputs( "<+>RamPressure()\n", debug_fp );
#	endif

	/* this is ram pressure - xMassDensity is density in gm/cm3,
	 * wind.windv is current wind velocity in cm/s */
	RamPressure_v = phycon.xMassDensity * POW2( wind.windv );

#	ifdef DEBUG_FUN
	fputs( " <->RamPressure()\n", debug_fp );
#	endif
	return( RamPressure_v );
}

