/*esccom fundamental escape probability radiative transfer routine, for complete redistribution */
/*escinc fundamental escape probability radiative transfer routine for incomplete redistribution */
/*escla escape prob for hydrogen atom Lya, using Hummer and Kunasz results,
 * called by hydropesc */
/*escsub escape probability radiative transfer for subordinate lines */
/*escgrd escape probability radiative transfer for incomplete redistribution */
/*esc_CRDwing escape probability for CRD with wings */
/*esc_CRDcore escape probability for CRD with no wings */
/*esca0k2 derive Hummer's K2 escape probability for Doppler core only */
/*eovrlp returns line destruction probability due to continuum opacity */

#include "cddefines.h"
#include "physconst.h"
#include "rt.h"
#include "hydrogenic.h"
#include "conrec.h"
#define	SCALE	2.
#include "converge.h"
#include "fhummr.h"
#include "rfield.h"
#include "chnukt.h"
#include "opacity.h"
#include "taulines.h"
#include "doppvel.h"
#include "rtescprob.h"

/*escmase escape probability for negative (masing) optical depths,*/
static double escmase(double tau);
/*humla fit Hummer and Kunasz escape probability for hydrogen atom Lya */
static void humla(double taume, 
  double beta, 
  float *esc, 
  float *dest,
  /* position of line in frequency array on c scale */
  long ipLine );

double escinc(double tau, 
  double a)
{
	double atau, 
	  b, 
	  escinc_v;

#	ifdef DEBUG_FUN
	fputs( "<+>escinc()\n", debug_fp );
#	endif


	/* this is one of the three fundamental escape probability routines
	 * the three are esccom, escinc, and escla
	 * it computes esc prob for incomplete redistribution
	 * */
	if( strcmp(rt.chEscFunSubord,"SIMP") == 0 )
	{
		/* this set with "escape simple" command, used for debugging */
		escinc_v = 1./(1. + tau);
		
#		ifdef DEBUG_FUN
		fputs( " <->escinc()\n", debug_fp );
#		endif
		return( escinc_v );
	}

	if( tau < 0. )
	{
		/* line mased */
		escinc_v = escmase(tau);
	}
	else if( tau < 10. )
	{
		/* linear part of doppler core */
		escinc_v = 1./(1. + 1.6*tau);
	}
	else
	{
		/* first find coeficient b(tau) */
		atau = a*tau;
		if( atau > 1. )
		{
			b = 1.6 + (3.*pow(2.*a,-0.12))/(1. + atau);
		}
		else
		{
			b = 1.6 + (3.*pow(2.*a,-0.12))/(1. + 1./sqrt(atau));
		}
		b = MIN2(6.,b);

		escinc_v = 1./(1. + b*tau);
	}

#	ifdef DEBUG_FUN
	fputs( " <->escinc()\n", debug_fp );
#	endif
	return( escinc_v );
}

/*esccom fundamental escape probability radiative transfer routine, for complete redistribution */
double esccom(double tau, 
  double a)
{
	double atau, 
	  b, 
	  esccom_v;

#	ifdef DEBUG_FUN
	fputs( "<+>esccom()\n", debug_fp );
#	endif

	/* this is one of the three fundamental escape probability routines
	 * the three are esccom, escinc, and escla
	 * it computes esc prob for complete redistribution
	 * computes escape prob for complete redistribution in one direction
	 * */
	if( tau < 0. )
	{
		/* line mased */
		esccom_v = escmase(tau);
	}

	else if( strcmp(rt.chEscFunSubord,"CRDW") == 0 )
	{
		/* this is the usual case for subordinate lines, 
		 * complete redistribution with damping wings */
		esccom_v = esca0k2(tau);
		if( tau > 1e3 )
		{
			esccom_v += 0.333*sqrt(a/(SQRTPI*tau));
		}
	}

	else if( strcmp(rt.chEscFunSubord,"SIMP") == 0 )
	{
		/* this set with "escape simple" command, used for debugging */
		esccom_v = 1./(1. + tau);
	}

	/* chEscFun is the escape prob function, usually k2 */
	else if( strcmp(rt.chEscFunSubord,"INCO") == 0 )
	{
		if( tau < 10. )
		{
			esccom_v = 1./(1. + 1.6*tau);
			
#			ifdef DEBUG_FUN
			fputs( " <->esccom()\n", debug_fp );
#			endif
			return( esccom_v );
		}

		/* first find coeficient b(tau) */
		atau = a*tau;
		if( atau > 1. )
		{
			b = 1.6 + (3.*pow(2.*a,-0.12))/(1. + atau);
		}
		else
		{
			b = 1.6 + (3.*pow(2.*a,-0.12))/(1. + 1./sqrt(atau));
		}

		b = MIN2(6.,b);
		esccom_v = 1./(1. + b*tau) + 0.25*sqrt(a/tau);

	}
	else if( strcmp(rt.chEscFunSubord,"  K2") == 0 )
	{
		/* this is the usual case, Hummer's k2 function */
		esccom_v = esca0k2(tau);
	}
	else
	{
		fprintf( ioQQQ, " Unknown escape probability function=%4.4s\n", 
		  rt.chEscFunSubord );
		puts( "[Stop in esccom]" );
		cdEXIT(1);exit(1);
	}

#	ifdef DEBUG_FUN
	fputs( " <->esccom()\n", debug_fp );
#	endif
	return( esccom_v );
}

/*escla escape prob for hydrogen atom Lya, using 
 >>refer	La	escp	Hummer, D.G., & Kunasz, P.B., 1980, ApJ, 236, 609
 * called by hydropesc */
double escla(float *esin, 
  float *dest, 
  double abund, 
  long int ipZ)/* 0 for H, */
{
	double beta, 
	  conopc, 
	  damp, 
	  escla_v;
	 float dstin, 
	  dstout;

#	ifdef DEBUG_FUN
	fputs( "<+>escla()\n", debug_fp );
#	endif

	/* 
	 * this is one of the three fundamental escape probability functions
	 * the three are esccom, escinc, and escla
	 * evaluate esc prob for LA
	 * optical depth in outer direction always defined
	 */

	/* check charge */
	assert( ipZ >= 0 && ipZ < LIMELM );

	if( HydroLines[ipZ][ipH2p][ipH1s].TauTot - 
	  HydroLines[ipZ][ipH2p][ipH1s].TauIn < 0. )
	{
		/* this is the case if we overrun the optical depth scale
		 * just leave things as they are */
		escla_v = HydroLines[ipZ][ipH2p][ipH1s].Pesc;
		rt.fracin = HydroLines[ipZ][ipH2p][ipH1s].FracInwd;
		*esin = rt.fracin;
		*dest = HydroLines[ipZ][ipH2p][ipH1s].Pdest;
		
#		ifdef DEBUG_FUN
		fputs( " <->escla()\n", debug_fp );
#		endif
		return( escla_v );
	}

	if( strcmp(rt.chRedisLa,"INCO") == 0 )
	{
		/* incomplete redistribution */
		conopc = opac.opac[HydroLines[ipZ][ipH2p][ipH1s].ipCont-1];
		if( abund > 0. )
		{
			/* the continuous opacity is positive, we have a valid soln */
			beta = conopc/(abund/SQRTPI*HydroLines[ipZ][ipH2p][ipH1s].opacity/
			  DoppVel.doppler[ipZ] + conopc);
		}
		else
		{
			/* abundance is zero, set miniumum dest prob */
			beta = 1e-10;
		}

		/* WAYIN will be escape prob in inward direction */
		humla(
		  HydroLines[ipZ][ipH2p][ipH1s].TauIn,
		  beta,
		  &rt.wayin,
		  &dstin , 
		  /* position of line in energy array on C scale */
		  HydroLines[ipZ][ipH2p][ipH1s].ipCont-1);

		if( (rt.wayin > 1.) || (rt.wayin < 0.) || (dstin > 1.) || (dstin < 0.) )
		{
			fprintf( ioQQQ, 
				" HUMLA returns inwd prob out of bounds;  esc=%10.2e dst="
				"%10.2e beta=%10.2e tau=%10.2e\n", 
			  rt.wayin, dstin, beta, HydroLines[ipZ][ipH2p][ipH1s].TauIn );
			puts( "[Stop in escla]" );
			cdEXIT(1);exit(1);
		}

		humla(MAX2(HydroLines[ipZ][ipH2p][ipH1s].TauTot/
		  100.,HydroLines[ipZ][ipH2p][ipH1s].TauTot-
		  HydroLines[ipZ][ipH2p][ipH1s].TauIn),
		  beta,&rt.wayout,&dstout, 
		  HydroLines[ipZ][ipH2p][ipH1s].ipCont-1);

		if( (rt.wayout > 1.) || (rt.wayout < 0.) || (dstout > 1.) || (dstout < 0.) )
		{
			fprintf( ioQQQ, 
				" HUMLA returns outwd prob out of bounds;  esc=%10.2e dst=%10.2e beta=%10.2e tau=%10.2e\n", 
			  rt.wayout, dstout, beta, 
			  MAX2(HydroLines[ipZ][ipH2p][ipH1s].TauTot/
			  100.,HydroLines[ipZ][ipH2p][ipH1s].TauTot-
			  HydroLines[ipZ][ipH2p][ipH1s].TauIn) );
			puts( "[Stop in escla]" );
			cdEXIT(1);exit(1);
		}

		escla_v = (rt.wayin + rt.wayout)/2.;
		*esin = rt.wayin;
		*dest = (dstin + dstout)/2.f;
		rt.fracin = rt.wayin/(rt.wayin + rt.wayout);
	}

	else if( (strcmp(rt.chRedisLa,"COMP") == 0 ) ||
		     (strcmp(rt.chRedisLa,"WIND") == 0 ) )
	{
		/* wind or complete redis */
		/*>>>chng 98 dec 18, set damp to zero before call, had not
		 * been initialized */
		damp = 0.;
		escla_v = escsub(HydroLines[ipZ][ipH2p][ipH1s].TauIn,
		  HydroLines[ipZ][ipH2p][ipH1s].TauTot,
		  damp);
		*esin = rt.wayin;
		rt.fracin = 0.5;
		*dest = 0.;
		/* not used, but keep lint quiet */
		beta = 1e-10;
	}

	else
	{
		/* not used, but keep lint quiet */
		beta = 1e-10;
		fprintf( ioQQQ,"escla called with insane redistribution function ==%s==\n",
			rt.chRedisLa );
		puts( "[Stop in escla]" );
		cdEXIT(1);exit(1);
	}

	assert( escla_v >= 0. );

	/* this is for debugging H Lya */
	{
		/*@-redef@*/
		enum {BUG=FALSE};
		/*@+redef@*/
		if( BUG )
		{
			if( ipZ == 0 )
			{
				fprintf(ioQQQ," %.2e %.2e %.2e %.2e %.2e \n",
				HydroLines[ipZ][ipH2p][ipH1s].TauIn,
				beta, 
				opac.albedo[HydroLines[ipZ][ipH2p][ipH1s].ipCont-1] ,
				escla_v , 
				*dest 
				);
			}
		}
	}

#	ifdef DEBUG_FUN
	fputs( " <->escla()\n", debug_fp );
#	endif
	return( escla_v );
}

/*escsub escape probability radiative transfer for subordinate lines */

double escsub(double tau, 
  double tout, 
  double damp)
{
	double escsub_v, 
	  t;

#	ifdef DEBUG_FUN
	fputs( "<+>escsub()\n", debug_fp );
#	endif

	if( opac.lgTauOutOn )
	{
		/* outward optical depths are defined, but don't let them go negat */
		t = tout - tau;
		/*  help convergence by not letting tau go to zero at back edge of
		 *  when there was a bad guess for the total optical depth
		 *note that this test is seldom hit since RTMakeStat does check for overrun */
		if( t < 0. )
			t = tau/SCALE;
		rt.wayin = (float)esccom(tau,damp);
		rt.wayout = (float)esccom(t,damp);
		rt.fracin = rt.wayin/(rt.wayin + rt.wayout);
		escsub_v = 0.5*(rt.wayin + rt.wayout);
	}
	else
	{
		/* first iteration, outward optical depths not defined
		 * dont estimate fraction in */
		rt.fracin = 0.;
		rt.wayout = 1.;
		escsub_v = esccom(tau,damp);
		rt.wayin = (float)escsub_v;
	}

#	ifdef DEBUG_FUN
	fputs( " <->escsub()\n", debug_fp );
#	endif
	return( escsub_v );
}

/*escgrd escape probability radiative transfer for incomplete redistribution */
double escgrd(double tau, 
  double tout, 
  double damp )
{
	double escgrd_v, 
	  tt;

#	ifdef DEBUG_FUN
	fputs( "<+>escgrd()\n", debug_fp );
#	endif

	/* find escape prob for incomp redis, average of two 1-sided probs*/

	if( opac.lgTauOutOn )
	{
		/*  outward optical depth if defined */
		tt = tout - tau;
		/*  help convergence by not letting tau go to zero at back edge of
		 *  when there was a bad guess for the total optical depth
		 *  note that this test is seldom hit since RTMakeStat does check
		 *  for overrun */
		if( tt < 0. )
		{
			tt = tau/SCALE;
		}

		rt.wayin = (float)escinc(tau,damp);
		rt.wayout = (float)escinc(tt,damp);
		rt.fracin = rt.wayin/(rt.wayin + rt.wayout);
		escgrd_v = 0.5*(rt.wayin + rt.wayout);
	}
	else
	{
		/*  outward optical depth not defined, dont estimate fraction out */
		rt.fracin = 0.;
		rt.wayout = 1.;
		escgrd_v = escinc(tau,damp);
		rt.wayin = (float)escgrd_v;
	}

	assert( escgrd_v > 0. );

#	ifdef DEBUG_FUN
	fputs( " <->escgrd()\n", debug_fp );
#	endif
	return( escgrd_v );
}

/*esc_CRDwing escape probability radiative transfer for CRDS in core only */
double esc_CRDwing(double tau, 
  double tout, 
  double damp)
{
	double escgrd_v, 
	  tt;

#	ifdef DEBUG_FUN
	fputs( "<+>esc_CRDwing()\n", debug_fp );
#	endif

	/* find escape prob for CRD with damping wings, average of two 1-sided probs*/

	/* crd with wings */
	if( opac.lgTauOutOn )
	{
		/*  outward optical depth if defined */
		tt = tout - tau;
		/*  help convergence by not letting tau go to zero at back edge of
		 *  when there was a bad guess for the total optical depth
		 *  note that this test is seldom hit since RTMakeStat does check
		 *  for overrun */
		if( tt < 0. )
		{
			tt = tau/SCALE;
		}

		rt.wayin = (float)esccom(tau,damp);
		rt.wayout = (float)esccom(tt,damp);
		rt.fracin = rt.wayin/(rt.wayin + rt.wayout);
		escgrd_v = 0.5*(rt.wayin + rt.wayout);
	}
	else
	{
		/*  outward optical depth not defined, dont estimate fraction out */
		rt.fracin = 0.;
		rt.wayout = 1.;
		escgrd_v = esccom(tau,damp);
		rt.wayin = (float)escgrd_v;
	}

	assert( escgrd_v > 0. );

#	ifdef DEBUG_FUN
	fputs( " <->esc_CRDwing()\n", debug_fp );
#	endif
	return( escgrd_v );
}

/*esc_CRDwing escape probability radiative transfer for incomplete redistribution */
double esc_CRDcore(double tau, 
  double tout)
{
	double escgrd_v, 
	  tt;

#	ifdef DEBUG_FUN
	fputs( "<+>esc_CRDcore()\n", debug_fp );
#	endif

	/* find escape prob for CRD with damping wings, average of two 1-sided probs*/

	/* crd with wings */
	if( opac.lgTauOutOn )
	{
		/*  outward optical depth if defined */
		tt = tout - tau;
		/*  help convergence by not letting tau go to zero at back edge of
		 *  when there was a bad guess for the total optical depth
		 *  note that this test is seldom hit since RTMakeStat does check
		 *  for overrun */
		if( tt < 0. )
		{
			tt = tau/SCALE;
		}

		rt.wayin = (float)esca0k2(tau);
		rt.wayout = (float)esca0k2(tt);
		rt.fracin = rt.wayin/(rt.wayin + rt.wayout);
		escgrd_v = 0.5*(rt.wayin + rt.wayout);
	}
	else
	{
		/*  outward optical depth not defined, dont estimate fraction out */
		rt.fracin = 0.;
		rt.wayout = 1.;
		escgrd_v = esca0k2(tau);
		rt.wayin = (float)escgrd_v;
	}

	assert( escgrd_v > 0. );

#	ifdef DEBUG_FUN
	fputs( " <->esc_CRDcore()\n", debug_fp );
#	endif
	return( escgrd_v );
}

/*esca0k2 derive Hummer's K2 escape probability for Doppler core only */
double esca0k2(double taume)
{
	double arg, 
	  esca0k2_v, 
	  suma, 
	  sumb, 
	  sumc, 
	  sumd, 
	  tau;
	static double a[5]={1.00,-0.1117897,-0.1249099917,-9.136358767e-3,
	  -3.370280896e-4};
	static double b[6]={1.00,0.1566124168,9.013261660e-3,1.908481163e-4,
	  -1.547417750e-7,-6.657439727e-9};
	static double c[5]={1.000,19.15049608,100.7986843,129.5307533,-31.43372468};
	static double d[6]={1.00,19.68910391,110.2576321,169.4911399,-16.69969409,
	  -36.664480000};

#	ifdef DEBUG_FUN
	fputs( "<+>esca0k2()\n", debug_fp );
#	endif

	/* compute Hummer's K2 escape probability function for a=0
	 * using approx from 
	 * >>refer	line	escp	Hummer, D.G., xxxx, JQRST, 26, 187.
	 *
	 * convert to David's opacity */
	tau = taume*SQRTPI;

	if( tau < 0. )
	{
		/* the line mased */
		esca0k2_v = escmase(taume);

	}
	else if( tau < 0.01 )
	{
		esca0k2_v = 1. - 2.*tau;

	}
	else if( tau <= 11. )
	{
		suma = a[0] + tau*(a[1] + tau*(a[2] + tau*(a[3] + a[4]*tau)));
		sumb = b[0] + tau*(b[1] + tau*(b[2] + tau*(b[3] + tau*(b[4] + 
		  b[5]*tau))));
		esca0k2_v = tau/2.5066283*log(tau/SQRTPI) + suma/sumb;

	}
	else
	{
		/* large optical depth limit */
		arg = 1./log(tau/SQRTPI);
		sumc = c[0] + arg*(c[1] + arg*(c[2] + arg*(c[3] + c[4]*arg)));
		sumd = d[0] + arg*(d[1] + arg*(d[2] + arg*(d[3] + arg*(d[4] + 
		  d[5]*arg))));
		esca0k2_v = (sumc/sumd)/(2.*tau*sqrt(log(tau/SQRTPI)));
	}


#	ifdef DEBUG_FUN
	fputs( " <->esca0k2()\n", debug_fp );
#	endif
	return( esca0k2_v );
}

/*escmase escape probability for negative (masing) optical depths */
void FindNeg( void )
{
	long int i;

#	ifdef DEBUG_FUN
	fputs( "<+>FindNeg()\n", debug_fp );
#	endif

	/* do the level 1 lines */
	for( i=1; i <= nLevel1; i++ )
	{
		/* check if a line was a strong maser */
		if( TauLines[i].TauIn < -1. )
			DumpLine(&TauLines[i]);
	}

	/* now do the level 2 lines */
	for( i=0; i < nWindLine; i++ )
	{
		/* check if a line was a strong maser */
		if( TauLine2[i].TauIn < -1. )
			DumpLine(&TauLine2[i]);
	}

	/* now do the co carbon monoxoide lines */
	for( i=0; i < nCORotate; i++ )
	{
		/* check if a line was a strong maser */
		if( C12O16Rotate[i].TauIn < -1. )
			DumpLine(&C12O16Rotate[i]);
		/* check if a line was a strong maser */
		if( C13O16Rotate[i].TauIn < -1. )
			DumpLine(&C13O16Rotate[i]);
	}

#	ifdef DEBUG_FUN
	fputs( " <->FindNeg()\n", debug_fp );
#	endif
	return;
}

static double escmase(double tau)
{
	double escmase_v;

#	ifdef DEBUG_FUN
	fputs( "<+>escmase()\n", debug_fp );
#	endif

	/* this is the only routine that computes maser escape proba */
	assert( tau <= 0. );

	if( tau > -0.1 )
	{
		escmase_v = 1. - tau*(0.5 + tau/6.);
	}
	else if( tau > -30. )
	{
		escmase_v = (1. - exp(-tau))/tau;
	}
	else
	{
		fprintf( ioQQQ, " escmase called with 2big tau%10.2e\n", 
		  tau  );
		fprintf( ioQQQ, " This is zone number%4ld\n", nzone );
		FindNeg();
		ShowMe();
		puts( "[Stop in escmase]" );
		cdEXIT(1);exit(1);
	}

	assert( escmase_v >= 1. );

#	ifdef DEBUG_FUN
	fputs( " <->escmase()\n", debug_fp );
#	endif
	return( escmase_v );
}

/*escConE2 one of the forms of the continuum escape probability */
/*cone2 generate e2 function needed for continuum transport */
double cone2(double t);

double escConE2(double x)
{
	double conesc_v;

#	ifdef DEBUG_FUN
	fputs( "<+>escConE2()\n", debug_fp );
#	endif

	conesc_v = exp(-chnukt.ContTkt*(x-1.))/
		x*cone2(chnukt.ctau/x/x/x);

#	ifdef DEBUG_FUN
	fputs( " <->escConE2()\n", debug_fp );
#	endif
	return( conesc_v );
}

/*cone2 generate e2 function needed for continuum transport */

double cone2(double t)
{
	double cone2_v, 
	  remain, 
	  tln;

#	ifdef DEBUG_FUN
	fputs( "<+>cone2()\n", debug_fp );
#	endif

	if( t < 80. )
	{
		tln = exp(-t);
	}
	else
	{
		cone2_v = 0.;
		
#		ifdef DEBUG_FUN
		fputs( " <->cone2()\n", debug_fp );
#		endif
		return( cone2_v );
	}

	/* fit of second exponential integral;
	 * T is optical depth, and TLN is EXP(-t)
	 * */
	if( t < 0.3 )
	{
		remain = (1.998069357 + t*(66.4037741 + t*107.2041376))/(1. + 
		  t*(37.4009646 + t*105.0388805));

	}
	else if( t < 20. )
	{
		remain = (1.823707708 + t*2.395042899)/(1. + t*(2.488885899 - 
		  t*0.00430538));

	}
	else
	{
		remain = 1.;
	}

	cone2_v = remain*tln/(2. + t);

#	ifdef DEBUG_FUN
	fputs( " <->cone2()\n", debug_fp );
#	endif
	return( cone2_v );
}

/*esccon continuum escape probability */

double esccon(double tau, 
  double hnukt)
{
	double dinc, 
	  escpcn_v, 
	  sumesc, 
	  sumrec;

#	ifdef DEBUG_FUN
	fputs( "<+>esccon()\n", debug_fp );
#	endif

	/* computes continuum escape probabilities */
	if( tau < 0.01 )
	{
		escpcn_v = 1.;
		
#		ifdef DEBUG_FUN
		fputs( " <->esccon()\n", debug_fp );
#		endif
		return( escpcn_v );
	}

	else if( hnukt > 1. && tau > 100. )
	{
		escpcn_v = 1e-20;
		
#		ifdef DEBUG_FUN
		fputs( " <->esccon()\n", debug_fp );
#		endif
		return( escpcn_v );
	}

	chnukt.ContTkt = (float)hnukt;
	chnukt.ctau = (float)tau;

	dinc = 10./hnukt;
	sumrec = qg32(1.,1.+dinc,conrec);
	sumesc = qg32(1.,1.+dinc,escConE2);

	if( sumrec > 0. )
	{
		escpcn_v = sumesc/sumrec;
	}
	else
	{
		escpcn_v = 0.;
	}

#	ifdef DEBUG_FUN
	fputs( " <->esccon()\n", debug_fp );
#	endif
	return( escpcn_v );
}

/*humla fit Hummer and Kunasz escape probability for hydrogen atom Lya */
static void humla(double taume, 
  double beta, 
  float *esc, 
  float *dest,
  /* position of line in frequency array on c scale */
  long ipLine )
{
	double esc0, 
	  fac, 
	  fac1, 
	  fac2, 
	  tau, 
	  taucon, 
	  taulog;

	/* DEST0 is the smallest destruction probability to return
	 * in high metallicity models */
	const double DEST0=1e-8 ;

#	ifdef DEBUG_FUN
	fputs( "<+>humla()\n", debug_fp );
#	endif

	/* fits to numerical results of Hummer and Kunasz Ap.J. 80 */
	tau = taume*SQRTPI;

	if( strcmp(rt.chEscFunSubord,"SIMP") == 0 )
	{
		/* this set with "escape simple" command, used for debugging */
		esc0 = 1./(1. + taume);
	}
	else
	{
		/* this is the real escape probability */
		esc0 = 1./((0.6451 + tau)*(0.47 + 1.08/(1. + 7.3e-6*tau)));
	}

	esc0 = MAX2(0.,esc0);
	esc0 = MIN2(1.,esc0);

	if( tau > 0. )
	{
		taulog = log10(MIN2(1e8,tau));
	}
	else
	{
		/* the line mased */
		hydro.xLaMase = MIN2(hydro.xLaMase,(float)tau);
		taulog = 0.;
		*dest = 0.;
		*esc = (float)esc0;
	}

	if( beta > 0. )
	{
		taucon = MIN2(2.,beta*tau);

		if( taucon > 1e-3 )
		{
			fac1 = -1.25 + 0.475*taulog;
			fac2 = -0.485 + 0.1615*taulog;
			fac = -fac1*taucon + fac2*POW2(taucon);
			fac = pow(10.,fac);
			fac = MIN2(1.,fac);
		}
		else
		{
			fac = 1.;
		}

		*esc = (float)(esc0*fac);
		/* MIN puts cat at 50 */
		*dest = (float)(beta/(0.30972 - MIN2(.28972,0.03541667*taulog)));
	}

	else
	{
		*dest = 0.;
		*esc = (float)esc0;
	}

	*dest = MIN2(*dest,1.f-*esc);
	*dest = MAX2(0.f,*dest);

	/* >>chng 99 apr 12, limit destruction prob in case where gas dominated by scattering.
	 * in this case scattering is much more likely than absorption on this event */
	*dest = (float)( (1. - opac.albedo[ipLine]) * *dest + 
		opac.albedo[ipLine]*DEST0);
	/* this is for debugging H Lya */
	{
		/*@-redef@*/
		enum {BUG=FALSE};
		/*@+redef@*/
		if( BUG )
		{
			if( ipLine < 600 )
			{
				fprintf(ioQQQ," %.2e %.2e %.2e %.2e %.2e \n",
				taume,
				beta, 
				(1. - opac.albedo[ipLine]), 
				opac.albedo[ipLine] ,
				*dest 
				);
			}
		}
	}

#	ifdef DEBUG_FUN
	fputs( " <->humla()\n", debug_fp );
#	endif
	return;
}

/*eovrlp returns line destruction probability due to continuum opacity */
double eovrlp(
	  /* abundance of species */
	  double abund, 
	  /* its line absorption cross section */
	  double crsec, 
	  /* pointer to energy within continuum array, to get background opacity,
	   * this is on the f not c scale */
	  long int ipanu, 
	  /* line width */
	  double widl, 
	  /* escape probability */
	  double escp, 
	  /* type of redistribution function */
	  char *chCore)
{
	/* this will be the value we shall return */
	double eovrlp_v;

	double conopc, 
	  beta;

	/* DEST0 is the smallest destruction probability to return
	 * in high metallicity models 
	 * this was set to 1e-8 until 99nov18,
	 * in cooling flow model the actual Lya ots dest prob was 1e-16,
	 * and this lower limit of 1e-8 caused energy balance problems,
	 * since dest prob was 8 orders of magnitude too great.  
	 * >>chng 99 nov 18, to 1e-20, but beware that comments indicate that
	 * this will cause problems with high metallicity clouds(?) */
	/* >>chng 00 jun 04, to 0 since very feeble ionization clouds, with almost zero opacity,
	 * this was a LARGE number */
	/*const double DEST0=1e-20 ;*/
	const double DEST0=0. ;

#	ifdef DEBUG_FUN
	fputs( "<+>eovrlp()\n", debug_fp );
#	endif

	/* computes "escape probability" due to continuum destruction of
	 *
	 * if esc prob gt 1 then line is masing - return small number for dest prob */
	/* >>>chng 99 apr 10, return min dest when scattering greater than abs */
	if( escp >= 1.0)
	{
		eovrlp_v = DEST0;
		
#		ifdef DEBUG_FUN
		fputs( " <->eovrlp()\n", debug_fp );
#		endif
		return( eovrlp_v );
	}

	/* find continuum opacity */
	conopc = opac.opac[ipanu-1];

	/* no idea of opacity whatsoever, on very first soln for this model */
	if( !conv.nTotalIoniz )
	{
		/* opacity not yet defined on very first try, return small but finite number */
		eovrlp_v = DEST0;
		
#		ifdef DEBUG_FUN
		fputs( " <->eovrlp()\n", debug_fp );
#		endif
		return( eovrlp_v );
	}

	/* may be no population, cannot use above test since return 0 not DEST0 */
	if( abund <= 0. || conopc <= 0. )
	{
		/* do not set this to DEST0 since energy not then conserved */
		eovrlp_v = 0.;
		
#		ifdef DEBUG_FUN
		fputs( " <->eovrlp()\n", debug_fp );
#		endif
		return( eovrlp_v );
	}

	/* fac of 1.7 convert to Hummer convention for line opacity */
	beta = conopc/(abund*SQRTPI*crsec/widl + conopc);
	beta = MIN2(beta,(1.-escp));

	if( strcmp(chCore,"INCO") == 0 )
	{
		/*  fits to 
		 *  >>>refer Hummer and Kunasz 1980 Ap.J. 236,609.
		 *  the max value of 1e-3 is so that we do not go too far
		 *  beyond what Hummer and Kunasz did */
		eovrlp_v = MIN2(1e-3,8.5*beta);
	}
	else if( strcmp(chCore,"  K2") == 0 )
	{
		/*  Doppler core only; a=0., Hummer 68 */
		eovrlp_v = fhummr(beta);
	}
	else if( strcmp(chCore,"SIMP") == 0 )
	{
		/*  this for debugging only */
		eovrlp_v = 8.5*beta;
	}
	else
	{
		fprintf( ioQQQ, " chCore of %4.4s not understood by EOVRLP.\n", 
		  chCore );
		puts( "[Stop in eovrlp]" );
		cdEXIT(1);exit(1);
	}

	eovrlp_v /= 1. + eovrlp_v;
	eovrlp_v *= 1. - escp;

	/*check results in bounds */
	assert( eovrlp_v >= 0.  );
	assert( eovrlp_v <= 1.  );

	{
		/* debugging code for Lya problems */
		/*@-redef@*/
		enum {DEBUG=FALSE};
		/*@+redef@*/
		if( DEBUG )
		{
			if( rfield.anu[ipanu-1]>0.73 && rfield.anu[ipanu-1]<0.76 &&
				(float)(abund)==HydroLines[0][ipH2p][ipH1s].PopOpc ) 
			{
				fprintf(ioQQQ,"%li eovrlp\t%g\n",
					nzone, eovrlp_v  );
			}
		}
	}

	/* >>chng 99 apr 12, limit destruction prob in case where gas dominated by scattering.
	 * in this case scattering is much more likely than absorption on this event */
	eovrlp_v = (1. - opac.albedo[ipanu-1]) * eovrlp_v + 
		opac.albedo[ipanu-1]*DEST0;

#	ifdef DEBUG_FUN
	fputs( " <->eovrlp()\n", debug_fp );
#	endif
	return( eovrlp_v );
}

